# $Id$
#
#  Copyright (C) 2003-2006  greg Landrum and Rational Discovery LLC
#
#   @@ All Rights Reserved @@
#  This file is part of the RDKit.
#  The contents are covered by the terms of the BSD license
#  which is included in the file license.txt, found at the root
#  of the RDKit source tree.
#
""" command line utility for growing composite models

**Usage**

  _GrowComposite [optional args] filename_

**Command Line Arguments**

  - -n *count*: number of new models to build

  - -C *pickle file name*:  name of file containing composite upon which to build.

  - --inNote *note*: note to be used in loading composite models from the database
      for growing

  - --balTable *table name*:  table from which to take the original data set
     (for balancing)

  - --balWeight *weight*: (between 0 and 1) weighting factor for the new data
     (for balancing). OR, *weight* can be a list of weights

  - --balCnt *count*: number of individual models in the balanced composite
     (for balancing)

  - --balH: use only the holdout set from the original data set in the balancing
     (for balancing)

  - --balT: use only the training set from the original data set in the balancing
     (for balancing)

  - -S: shuffle the original data set
     (for balancing)

  - -r: randomize the activities of the original data set
     (for balancing)

  - -N *note*: note to be attached to the grown composite when it's saved in the
     database

  - --outNote *note*: equivalent to -N

  - -o *filename*: name of an output file to hold the pickled composite after
     it has been grown.
     If multiple balance weights are used, the weights will be added to
     the filenames.

  - -L *limit*: provide an (integer) limit on individual model complexity

  - -d *database name*: instead of reading the data from a QDAT file,
     pull it from a database.  In this case, the _filename_ argument
     provides the name of the database table containing the data set.

  - -p *tablename*: store persistence data in the database
     in table *tablename*

  - -l: locks the random number generator to give consistent sets
     of training and hold-out data.  This is primarily intended
     for testing purposes.

  - -g: be less greedy when training the models.

  - -G *number*: force trees to be rooted at descriptor *number*.

  - -D: show a detailed breakdown of the composite model performance
     across the training and, when appropriate, hold-out sets.

  - -t *threshold value*: use high-confidence predictions for the final
     analysis of the hold-out data.

  - -q *list string*:  Add QuantTrees to the composite and use the list
     specified in *list string* as the number of target quantization
     bounds for each descriptor.  Don't forget to include 0's at the
     beginning and end of *list string* for the name and value fields.
     For example, if there are 4 descriptors and you want 2 quant bounds
     apiece, you would use _-q "[0,2,2,2,2,0]"_.
     Two special cases:
       1) If you would like to ignore a descriptor in the model building,
          use '-1' for its number of quant bounds.
       2) If you have integer valued data that should not be quantized
          further, enter 0 for that descriptor.

  - -V: print the version number and exit

"""
from warnings import warn

warn('This module is deprecated and will be removed in the 2024.03 release', DeprecationWarning,
     stacklevel=2)

import pickle
import sys
import time

import numpy

from rdkit.Dbase.DbConnection import DbConnect
from rdkit.ML import BuildComposite, CompositeRun, ScreenComposite
from rdkit.ML.Composite import AdjustComposite
from rdkit.ML.Data import DataUtils, SplitData

_runDetails = CompositeRun.CompositeRun()

__VERSION_STRING = "0.5.0"

_verbose = 1


def message(msg):
  """ emits messages to _sys.stdout_
    override this in modules which import this one to redirect output

    **Arguments**

      - msg: the string to be displayed

  """
  if _verbose:
    sys.stdout.write('%s\n' % (msg))


def GrowIt(details, composite, progressCallback=None, saveIt=1, setDescNames=0, data=None):
  """ does the actual work of building a composite model

    **Arguments**

      - details:  a _CompositeRun.CompositeRun_ object containing details
        (options, parameters, etc.) about the run

      - composite: the composite model to grow

      - progressCallback: (optional) a function which is called with a single
        argument (the number of models built so far) after each model is built.

      - saveIt: (optional) if this is nonzero, the resulting model will be pickled
        and dumped to the filename specified in _details.outName_

      - setDescNames: (optional) if nonzero, the composite's _SetInputOrder()_ method
        will be called using the results of the data set's _GetVarNames()_ method;
        it is assumed that the details object has a _descNames attribute which
        is passed to the composites _SetDescriptorNames()_ method.  Otherwise
        (the default), _SetDescriptorNames()_ gets the results of _GetVarNames()_.

      - data: (optional) the data set to be used.  If this is not provided, the
        data set described in details will be used.

    **Returns**

      the enlarged composite model


  """
  details.rundate = time.asctime()

  if data is None:
    fName = details.tableName.strip()
    if details.outName == '':
      details.outName = fName + '.pkl'
    if details.dbName == '':
      data = DataUtils.BuildQuantDataSet(fName)
    elif details.qBounds != []:
      details.tableName = fName
      data = details.GetDataSet()
    else:
      data = DataUtils.DBToQuantData(  # Function no longer defined
        details.dbName, fName, quantName=details.qTableName, user=details.dbUser,
        password=details.dbPassword)

  seed = composite._randomSeed
  DataUtils.InitRandomNumbers(seed)
  if details.shuffleActivities == 1:
    DataUtils.RandomizeActivities(data, shuffle=1, runDetails=details)
  elif details.randomActivities == 1:
    DataUtils.RandomizeActivities(data, shuffle=0, runDetails=details)

  namedExamples = data.GetNamedData()
  trainExamples = namedExamples
  nExamples = len(trainExamples)
  message('Training with %d examples' % (nExamples))
  message('\t%d descriptors' % (len(trainExamples[0]) - 2))
  nVars = data.GetNVars()
  nPossibleVals = composite.nPossibleVals
  attrs = list(range(1, nVars + 1))

  if details.useTrees:
    from rdkit.ML.DecTree import CrossValidate, PruneTree
    if details.qBounds != []:
      from rdkit.ML.DecTree import BuildQuantTree
      builder = BuildQuantTree.QuantTreeBoot
    else:
      from rdkit.ML.DecTree import ID3
      builder = ID3.ID3Boot
    driver = CrossValidate.CrossValidationDriver
    pruner = PruneTree.PruneTree

    if setDescNames:
      composite.SetInputOrder(data.GetVarNames())
    composite.Grow(trainExamples, attrs, [0] + nPossibleVals, buildDriver=driver, pruner=pruner,
                   nTries=details.nModels, pruneIt=details.pruneIt, lessGreedy=details.lessGreedy,
                   needsQuantization=0, treeBuilder=builder, nQuantBounds=details.qBounds,
                   startAt=details.startAt, maxDepth=details.limitDepth,
                   progressCallback=progressCallback, silent=not _verbose)

  else:
    from rdkit.ML.Neural import CrossValidate
    driver = CrossValidate.CrossValidationDriver
    composite.Grow(trainExamples, attrs, [0] + nPossibleVals, nTries=details.nModels,
                   buildDriver=driver, needsQuantization=0)

  composite.AverageErrors()
  composite.SortModels()
  modelList, counts, avgErrs = composite.GetAllData()
  counts = numpy.array(counts)
  avgErrs = numpy.array(avgErrs)
  composite._varNames = data.GetVarNames()

  for i in range(len(modelList)):
    modelList[i].NameModel(composite._varNames)

  # do final statistics
  weightedErrs = counts * avgErrs
  averageErr = sum(weightedErrs) / sum(counts)
  devs = (avgErrs - averageErr)
  devs = devs * counts
  devs = numpy.sqrt(devs * devs)
  avgDev = sum(devs) / sum(counts)
  if _verbose:
    message('# Overall Average Error: %%% 5.2f, Average Deviation: %%% 6.2f' %
            (100. * averageErr, 100. * avgDev))

  if details.bayesModel:
    composite.Train(trainExamples, verbose=0)

  badExamples = []
  if not details.detailedRes:
    if _verbose:
      message('Testing all examples')
    wrong = BuildComposite.testall(composite, namedExamples, badExamples)
    if _verbose:
      message('%d examples (%% %5.2f) were misclassified' %
              (len(wrong), 100. * float(len(wrong)) / float(len(namedExamples))))
    _runDetails.overall_error = float(len(wrong)) / len(namedExamples)

  if details.detailedRes:
    if _verbose:
      message('\nEntire data set:')
    resTup = ScreenComposite.ShowVoteResults(range(data.GetNPts()), data, composite,
                                             nPossibleVals[-1], details.threshold)
    nGood, nBad, _, avgGood, avgBad, _, voteTab = resTup
    nPts = len(namedExamples)
    nClass = nGood + nBad
    _runDetails.overall_error = float(nBad) / nClass
    _runDetails.overall_correct_conf = avgGood
    _runDetails.overall_incorrect_conf = avgBad
    _runDetails.overall_result_matrix = repr(voteTab)
    nRej = nClass - nPts
    if nRej > 0:
      _runDetails.overall_fraction_dropped = float(nRej) / nPts

  return composite


def GetComposites(details):
  res = []
  if details.persistTblName and details.inNote:
    conn = DbConnect(details.dbName, details.persistTblName)
    mdls = conn.GetData(fields='MODEL', where="where note='%s'" % (details.inNote))
    for row in mdls:
      rawD = row[0]
      res.append(pickle.loads(str(rawD)))
  elif details.composFileName:
    res.append(pickle.load(open(details.composFileName, 'rb')))
  return res


def BalanceComposite(details, composite, data1=None, data2=None):
  """ balances the composite using the parameters provided in details

   **Arguments**

     - details a _CompositeRun.RunDetails_ object

     - composite: the composite model to be balanced

     - data1: (optional) if provided, this should be the
       data set used to construct the original models

     - data2: (optional) if provided, this should be the
       data set used to construct the new individual models

  """
  if not details.balCnt or details.balCnt > len(composite):
    return composite
  message("Balancing Composite")

  #
  # start by getting data set 1: which is the data set used to build the
  #  original models
  #
  if data1 is None:
    message("\tReading First Data Set")
    fName = details.balTable.strip()
    tmp = details.tableName
    details.tableName = fName
    dbName = details.dbName
    details.dbName = details.balDb
    data1 = details.GetDataSet()
    details.tableName = tmp
    details.dbName = dbName
  if data1 is None:
    return composite
  details.splitFrac = composite._splitFrac
  details.randomSeed = composite._randomSeed
  DataUtils.InitRandomNumbers(details.randomSeed)
  if details.shuffleActivities == 1:
    DataUtils.RandomizeActivities(data1, shuffle=1, runDetails=details)
  elif details.randomActivities == 1:
    DataUtils.RandomizeActivities(data1, shuffle=0, runDetails=details)
  namedExamples = data1.GetNamedData()
  if details.balDoHoldout or details.balDoTrain:
    trainIdx, testIdx = SplitData.SplitIndices(len(namedExamples), details.splitFrac, silent=1)
    trainExamples = [namedExamples[x] for x in trainIdx]
    testExamples = [namedExamples[x] for x in testIdx]
    if details.filterFrac != 0.0:
      trainIdx, temp = DataUtils.FilterData(trainExamples, details.filterVal, details.filterFrac,
                                            -1, indicesOnly=1)
      tmp = [trainExamples[x] for x in trainIdx]
      testExamples += [trainExamples[x] for x in temp]
      trainExamples = tmp
    if details.balDoHoldout:
      testExamples, trainExamples = trainExamples, testExamples
  else:
    trainExamples = namedExamples
  dataSet1 = trainExamples
  cols1 = [x.upper() for x in data1.GetVarNames()]
  data1 = None

  #
  # now grab data set 2: the data used to build the new individual models
  #
  if data2 is None:
    message("\tReading Second Data Set")
    data2 = details.GetDataSet()
  if data2 is None:
    return composite
  details.splitFrac = composite._splitFrac
  details.randomSeed = composite._randomSeed
  DataUtils.InitRandomNumbers(details.randomSeed)
  if details.shuffleActivities == 1:
    DataUtils.RandomizeActivities(data2, shuffle=1, runDetails=details)
  elif details.randomActivities == 1:
    DataUtils.RandomizeActivities(data2, shuffle=0, runDetails=details)
  dataSet2 = data2.GetNamedData()
  cols2 = [x.upper() for x in data2.GetVarNames()]
  data2 = None

  # and balance it:
  res = []
  weights = details.balWeight
  if not isinstance(weights, (tuple, list)):
    weights = (weights, )
  for weight in weights:
    message("\tBalancing with Weight: %.4f" % (weight))
    res.append(
      AdjustComposite.BalanceComposite(composite, dataSet1, dataSet2, weight, details.balCnt,
                                       names1=cols1, names2=cols2))
  return res


def ShowVersion(includeArgs=0):
  """ prints the version number

  """
  print('This is GrowComposite.py version %s' % (__VERSION_STRING))
  if includeArgs:
    print('command line was:')
    print(' '.join(sys.argv))


def Usage():
  """ provides a list of arguments for when this is used from the command line

  """
  print(__doc__)
  sys.exit(-1)


def SetDefaults(runDetails=None):
  """  initializes a details object with default values

      **Arguments**

        - details:  (optional) a _CompositeRun.CompositeRun_ object.
          If this is not provided, the global _runDetails will be used.

      **Returns**

        the initialized _CompositeRun_ object.


  """
  if runDetails is None:
    runDetails = _runDetails
  return CompositeRun.SetDefaults(runDetails)


def ParseArgs(runDetails):
  """ parses command line arguments and updates _runDetails_

      **Arguments**

        - runDetails:  a _CompositeRun.CompositeRun_ object.

  """
  import getopt
  args, extra = getopt.getopt(sys.argv[1:], 'P:o:n:p:b:sf:F:v:hlgd:rSTt:Q:q:DVG:L:C:N:', [
    'inNote=',
    'outNote=',
    'balTable=',
    'balWeight=',
    'balCnt=',
    'balH',
    'balT',
    'balDb=',
  ])
  runDetails.inNote = ''
  runDetails.composFileName = ''
  runDetails.balTable = ''
  runDetails.balWeight = (0.5, )
  runDetails.balCnt = 0
  runDetails.balDoHoldout = 0
  runDetails.balDoTrain = 0
  runDetails.balDb = ''
  for arg, val in args:
    if arg == '-n':
      runDetails.nModels = int(val)
    elif arg == '-C':
      runDetails.composFileName = val
    elif arg == '--balTable':
      runDetails.balTable = val
    elif arg == '--balWeight':
      runDetails.balWeight = eval(val)
      if not isinstance(runDetails.balWeight, (tuple, list)):
        runDetails.balWeight = (runDetails.balWeight, )
    elif arg == '--balCnt':
      runDetails.balCnt = int(val)
    elif arg == '--balH':
      runDetails.balDoHoldout = 1
    elif arg == '--balT':
      runDetails.balDoTrain = 1
    elif arg == '--balDb':
      runDetails.balDb = val
    elif arg == '--inNote':
      runDetails.inNote = val
    elif arg == '-N' or arg == '--outNote':
      runDetails.note = val
    elif arg == '-o':
      runDetails.outName = val
    elif arg == '-p':
      runDetails.persistTblName = val
    elif arg == '-r':
      runDetails.randomActivities = 1
    elif arg == '-S':
      runDetails.shuffleActivities = 1
    elif arg == '-h':
      Usage()
    elif arg == '-l':
      runDetails.lockRandom = 1
    elif arg == '-g':
      runDetails.lessGreedy = 1
    elif arg == '-G':
      runDetails.startAt = int(val)
    elif arg == '-d':
      runDetails.dbName = val
    elif arg == '-T':
      runDetails.useTrees = 0
    elif arg == '-t':
      runDetails.threshold = float(val)
    elif arg == '-D':
      runDetails.detailedRes = 1
    elif arg == '-L':
      runDetails.limitDepth = int(val)
    elif arg == '-q':
      qBounds = eval(val)
      assert isinstance(qBounds,
                        (tuple, list)), 'bad argument type for -q, specify a list as a string'
      runDetails.qBoundCount = val
      runDetails.qBounds = qBounds
    elif arg == '-Q':
      qBounds = eval(val)
      assert type(qBounds) in [type([]), type(
        ())], 'bad argument type for -Q, specify a list as a string'
      runDetails.activityBounds = qBounds
      runDetails.activityBoundsVals = val
    elif arg == '-V':
      ShowVersion()
      sys.exit(0)
    else:
      print('bad argument:', arg, file=sys.stderr)
      Usage()
  runDetails.tableName = extra[0]
  if not runDetails.balDb:
    runDetails.balDb = runDetails.dbName


if __name__ == '__main__':
  if len(sys.argv) < 2:
    Usage()

  _runDetails.cmd = ' '.join(sys.argv)
  SetDefaults(_runDetails)
  ParseArgs(_runDetails)

  ShowVersion(includeArgs=1)

  initModels = GetComposites(_runDetails)
  nModels = len(initModels)
  if nModels > 1:
    for i in range(nModels):
      sys.stderr.write(
        '---------------------------------\n\tDoing %d of %d\n---------------------------------\n' %
        (i + 1, nModels))
      composite = GrowIt(_runDetails, initModels[i], setDescNames=1)
      if _runDetails.balTable and _runDetails.balCnt:
        composites = BalanceComposite(_runDetails, composite)
      else:
        composites = [composite]
      for mdl in composites:
        mdl.ClearModelExamples()
      if _runDetails.outName:
        nWeights = len(_runDetails.balWeight)
        if nWeights == 1:
          outName = _runDetails.outName
          composites[0].Pickle(outName)
        else:
          for i in range(nWeights):
            weight = int(100 * _runDetails.balWeight[i])
            model = composites[i]
            outName = '%s.%d.pkl' % (_runDetails.outName.split('.pkl')[0], weight)
            model.Pickle(outName)
      if _runDetails.persistTblName and _runDetails.dbName:
        message('Updating results table %s:%s' % (_runDetails.dbName, _runDetails.persistTblName))
        if (len(_runDetails.balWeight)) > 1:
          message('WARNING: updating results table with models having different weights')
        # save the composite
        for i in range(len(composites)):
          _runDetails.model = pickle.dumps(composites[i])
          _runDetails.Store(db=_runDetails.dbName, table=_runDetails.persistTblName)
  elif nModels == 1:
    composite = GrowIt(_runDetails, initModels[0], setDescNames=1)
    if _runDetails.balTable and _runDetails.balCnt:
      composites = BalanceComposite(_runDetails, composite)
    else:
      composites = [composite]
    for mdl in composites:
      mdl.ClearModelExamples()
    if _runDetails.outName:
      nWeights = len(_runDetails.balWeight)
      if nWeights == 1:
        outName = _runDetails.outName
        composites[0].Pickle(outName)
      else:
        for i in range(nWeights):
          weight = int(100 * _runDetails.balWeight[i])
          model = composites[i]
          outName = '%s.%d.pkl' % (_runDetails.outName.split('.pkl')[0], weight)
          model.Pickle(outName)
    if _runDetails.persistTblName and _runDetails.dbName:
      message('Updating results table %s:%s' % (_runDetails.dbName, _runDetails.persistTblName))
      if (len(composites)) > 1:
        message('WARNING: updating results table with models having different weights')
      for i in range(len(composites)):
        _runDetails.model = pickle.dumps(composites[i])
        _runDetails.Store(db=_runDetails.dbName, table=_runDetails.persistTblName)
  else:
    message("No models found")



# $Id$
#
#  Copyright (C) 2003-2008  greg Landrum and Rational Discovery LLC
#
#   @@ All Rights Reserved @@
#  This file is part of the RDKit.
#  The contents are covered by the terms of the BSD license
#  which is included in the file license.txt, found at the root
#  of the RDKit source tree.
#
"""unit testing code for the ScreenComposite functionality

"""
import io
import os
import pickle
import unittest

from rdkit import RDConfig
from rdkit.ML import ScreenComposite


class TestCase(unittest.TestCase):

  def setUp(self):
    self.baseDir = os.path.join(RDConfig.RDCodeDir, 'ML', 'test_data')
    self.dbName = RDConfig.RDTestDatabase
    self.details = ScreenComposite.SetDefaults()
    self.details.dbName = self.dbName
    self.details.dbUser = RDConfig.defaultDBUser
    self.details.dbPassword = RDConfig.defaultDBPassword

  def test1_basics(self):
    # """ basics """
    self.details.tableName = 'ferro_quant'
    with open(os.path.join(self.baseDir, 'ferromag_quant_10.pkl'), 'r') as pklTF:
      buf = pklTF.read().replace('\r\n', '\n').encode('utf-8')
      pklTF.close()
    with io.BytesIO(buf) as pklF:
      compos = pickle.load(pklF)
    tgt = 5
    self.assertEqual(len(compos), tgt)

    nGood, misCount, nSkipped, avgGood, avgBad, avgSkip, tbl = ScreenComposite.ScreenFromDetails(
      compos, self.details)
    self.assertEqual(nGood, 93)
    self.assertEqual(misCount, 2)
    self.assertEqual(nSkipped, 0)
    self.assertAlmostEqual(avgGood, .9871, 4)
    self.assertAlmostEqual(avgBad, .8000, 4)
    self.assertAlmostEqual(avgSkip, 0, 4)
    self.assertEqual(tbl[0, 0], 54)
    self.assertEqual(tbl[1, 1], 39)
    self.assertEqual(tbl[0, 1], 2)
    self.assertEqual(tbl[1, 0], 0)

  def test2_include_holdout(self):
    # """ include holdout data only """
    self.details.tableName = 'ferro_quant'
    self.details.doHoldout = 1
    self.details.doTraining = 0

    with open(os.path.join(self.baseDir, 'ferromag_quant_10.pkl'), 'r') as pklTF:
      buf = pklTF.read().replace('\r\n', '\n').encode('utf-8')
      pklTF.close()
    with io.BytesIO(buf) as pklF:
      compos = pickle.load(pklF)
    tgt = 5
    self.assertEqual(len(compos), tgt)

    nGood, misCount, nSkipped, avgGood, avgBad, avgSkip, tbl = ScreenComposite.ScreenFromDetails(
      compos, self.details)
    self.assertEqual(nGood, 28)
    self.assertEqual(misCount, 1)
    self.assertEqual(nSkipped, 0)
    self.assertAlmostEqual(avgGood, .9964, 4)
    self.assertAlmostEqual(avgBad, 1.000, 4)
    self.assertAlmostEqual(avgSkip, 0, 4)
    self.assertEqual(tbl[0, 0], 16)
    self.assertEqual(tbl[1, 1], 12)
    self.assertEqual(tbl[0, 1], 1)
    self.assertEqual(tbl[1, 0], 0)

  def test3_include_training(self):
    # """ include training data only """
    self.details.tableName = 'ferro_quant'
    self.details.doHoldout = 0
    self.details.doTraining = 1

    with open(os.path.join(self.baseDir, 'ferromag_quant_10.pkl'), 'r') as pklTF:
      buf = pklTF.read().replace('\r\n', '\n').encode('utf-8')
      pklTF.close()
    with io.BytesIO(buf) as pklF:
      compos = pickle.load(pklF)
    tgt = 5
    self.assertEqual(len(compos), tgt, 'bad composite loaded: %d != %d' % (len(compos), tgt))

    nGood, misCount, nSkipped, avgGood, avgBad, avgSkip, tbl = ScreenComposite.ScreenFromDetails(
      compos, self.details)
    self.assertEqual(nGood, 65)
    self.assertEqual(misCount, 1)
    self.assertEqual(nSkipped, 0)
    self.assertAlmostEqual(avgGood, .98307, 4)
    self.assertAlmostEqual(avgBad, 0.600, 4)
    self.assertAlmostEqual(avgSkip, 0, 4)
    self.assertEqual(tbl[0, 0], 38, tbl)
    self.assertEqual(tbl[1, 1], 27)
    self.assertEqual(tbl[0, 1], 1)
    self.assertEqual(tbl[1, 0], 0)

  def test4_thresholding(self):
    # """ include thresholding """
    self.details.tableName = 'ferro_quant'
    self.details.threshold = 0.80
    self.details.doHoldout = 0
    self.details.doTraining = 0

    with open(os.path.join(self.baseDir, 'ferromag_quant_10.pkl'), 'r') as pklTF:
      buf = pklTF.read().replace('\r\n', '\n').encode('utf-8')
      pklTF.close()
    with io.BytesIO(buf) as pklF:
      compos = pickle.load(pklF)
    tgt = 5
    self.assertEqual(len(compos), tgt)

    nGood, misCount, nSkipped, avgGood, avgBad, avgSkip, tbl = ScreenComposite.ScreenFromDetails(
      compos, self.details)
    self.assertEqual(nGood, 91)
    self.assertEqual(misCount, 1)
    self.assertEqual(nSkipped, 3)
    self.assertAlmostEqual(avgGood, 0.9956, 4)
    self.assertAlmostEqual(avgBad, 1.000, 4)
    self.assertAlmostEqual(avgSkip, 0.6000, 4)
    self.assertEqual(tbl[0, 0], 54)
    self.assertEqual(tbl[1, 1], 37)
    self.assertEqual(tbl[0, 1], 1)
    self.assertEqual(tbl[1, 0], 0)

  def test5_basics(self):
    # """ basics """
    self.details.tableName = 'ferro_noquant'

    with open(os.path.join(self.baseDir, 'ferromag_auto_10_3.pkl'), 'r') as pklTF:
      buf = pklTF.read().replace('\r\n', '\n').encode('utf-8')
      pklTF.close()
    with io.BytesIO(buf) as pklF:
      compos = pickle.load(pklF)
    tgt = 10
    self.assertEqual(len(compos), tgt)

    tpl = ScreenComposite.ScreenFromDetails(compos, self.details)
    nGood, misCount, nSkipped, avgGood, avgBad, avgSkip, tbl = tpl

    self.assertEqual(nGood, 95)
    self.assertEqual(misCount, 8)
    self.assertEqual(nSkipped, 0)
    self.assertAlmostEqual(avgGood, .9684, 4)
    self.assertAlmostEqual(avgBad, .8375, 4)
    self.assertAlmostEqual(avgSkip, 0, 4)
    self.assertEqual(tbl[0, 0], 50)
    self.assertEqual(tbl[1, 1], 45)
    self.assertEqual(tbl[0, 1], 5)
    self.assertEqual(tbl[1, 0], 3)

  def test6_multiple_models(self):
    # """ multiple models """
    self.details.tableName = 'ferro_noquant'
    with open(os.path.join(self.baseDir, 'ferromag_auto_10_3.pkl'), 'r') as pklTF:
      buf = pklTF.read().replace('\r\n', '\n').encode('utf-8')
      pklTF.close()
    with io.BytesIO(buf) as pklF:
      compos = pickle.load(pklF)
    tgt = 10
    self.assertEqual(len(compos), tgt)
    composites = [compos, compos]
    tpl = ScreenComposite.ScreenFromDetails(composites, self.details)
    nGood, misCount, nSkipped, avgGood, avgBad, avgSkip, tbl = tpl
    self.assertEqual(nGood[0], 95)
    self.assertEqual(misCount[0], 8)
    self.assertEqual(nSkipped[0], 0)
    self.assertAlmostEqual(avgGood[0], .9684, 4)
    self.assertAlmostEqual(avgBad[0], .8375, 4)
    self.assertAlmostEqual(avgSkip[0], 0.0, 4)
    self.assertEqual(nGood[1], 0)
    self.assertEqual(misCount[1], 0)
    self.assertEqual(nSkipped[1], 0)
    self.assertEqual(avgGood[1], 0)
    self.assertEqual(avgBad[1], 0)
    self.assertEqual(avgSkip[1], 0)
    self.assertEqual(tbl[0, 0], 50)
    self.assertEqual(tbl[1, 1], 45)
    self.assertEqual(tbl[0, 1], 5)
    self.assertEqual(tbl[1, 0], 3)

  def test7_shuffle(self):
    # """ shuffle """
    self.details.tableName = 'ferro_noquant'
    with open(os.path.join(self.baseDir, 'ferromag_shuffle_10_3.pkl'), 'r') as pklTF:
      buf = pklTF.read().replace('\r\n', '\n').encode('utf-8')
      pklTF.close()
    with io.BytesIO(buf) as pklF:
      compos = pickle.load(pklF)
    tgt = 10
    self.assertEqual(len(compos), tgt)
    self.details.shuffleActivities = 1
    nGood, misCount, nSkipped, avgGood, avgBad, avgSkip, tbl = ScreenComposite.ScreenFromDetails(
      compos, self.details)
    self.assertEqual(nGood, 50)
    self.assertEqual(misCount, 53)
    self.assertEqual(nSkipped, 0)
    self.assertAlmostEqual(avgGood, .7380, 4)
    self.assertAlmostEqual(avgBad, .7660, 4)
    self.assertAlmostEqual(avgSkip, 0, 4)
    self.assertEqual(tbl[0, 0], 30)
    self.assertEqual(tbl[1, 1], 20)
    self.assertEqual(tbl[0, 1], 25)
    self.assertEqual(tbl[1, 0], 28)

  def test8_shuffle_segmentation(self):
    # """ shuffle with segmentation """
    self.details.tableName = 'ferro_noquant'
    with open(os.path.join(self.baseDir, 'ferromag_shuffle_10_3.pkl'), 'r') as pklTF:
      buf = pklTF.read().replace('\r\n', '\n').encode('utf-8')
      pklTF.close()
    with io.BytesIO(buf) as pklF:
      compos = pickle.load(pklF)
    tgt = 10
    self.assertEqual(len(compos), tgt)
    self.details.shuffleActivities = 1
    self.details.doHoldout = 1
    nGood, misCount, nSkipped, avgGood, avgBad, avgSkip, tbl = ScreenComposite.ScreenFromDetails(
      compos, self.details)
    self.assertEqual(nGood, 19)
    self.assertEqual(misCount, 12)
    self.assertEqual(nSkipped, 0)
    self.assertAlmostEqual(avgGood, .7737, 4)
    self.assertAlmostEqual(avgBad, .7500, 4)
    self.assertAlmostEqual(avgSkip, 0, 4)
    self.assertEqual(tbl[0, 0], 12)
    self.assertEqual(tbl[1, 1], 7)
    self.assertEqual(tbl[0, 1], 6)
    self.assertEqual(tbl[1, 0], 6)

  def test9_shuffle_segmentation2(self):
    # """ shuffle with segmentation2 """
    self.details.tableName = 'ferro_noquant'
    with open(os.path.join(self.baseDir, 'ferromag_shuffle_10_3.pkl'), 'r') as pklTF:
      buf = pklTF.read().replace('\r\n', '\n').encode('utf-8')
      pklTF.close()
    with io.BytesIO(buf) as pklF:
      compos = pickle.load(pklF)
    tgt = 10
    self.assertEqual(len(compos), tgt)
    self.details.shuffleActivities = 1
    self.details.doTraining = 1
    nGood, misCount, nSkipped, avgGood, avgBad, avgSkip, tbl = ScreenComposite.ScreenFromDetails(
      compos, self.details)
    self.assertEqual(nGood, 31)
    self.assertEqual(misCount, 41)
    self.assertEqual(nSkipped, 0)
    self.assertAlmostEqual(avgGood, .7161, 4)
    self.assertAlmostEqual(avgBad, .7707, 4)
    self.assertAlmostEqual(avgSkip, 0.0, 4)
    self.assertEqual(tbl[0, 0], 18)
    self.assertEqual(tbl[1, 1], 13)
    self.assertEqual(tbl[0, 1], 19)
    self.assertEqual(tbl[1, 0], 22)

  def test10_filtering(self):
    # """ filtering """
    self.details.tableName = 'ferro_noquant'
    with open(os.path.join(self.baseDir, 'ferromag_filt_10_3.pkl'), 'r') as pklTF:
      buf = pklTF.read().replace('\r\n', '\n').encode('utf-8')
      pklTF.close()
    with io.BytesIO(buf) as pklF:
      compos = pickle.load(pklF)
    tgt = 10
    self.assertEqual(len(compos), tgt)
    self.details.filterVal = 1
    self.details.filterFrac = .33

    nGood, misCount, nSkipped, avgGood, avgBad, avgSkip, tbl = ScreenComposite.ScreenFromDetails(
      compos, self.details)
    self.assertEqual(nGood, 90)
    self.assertEqual(misCount, 13)
    self.assertEqual(nSkipped, 0)
    self.assertAlmostEqual(avgGood, .9578, 4)
    self.assertAlmostEqual(avgBad, .8538, 4)
    self.assertAlmostEqual(avgSkip, 0, 4)
    self.assertEqual(tbl[0, 0], 54)
    self.assertEqual(tbl[1, 1], 36)
    self.assertEqual(tbl[0, 1], 1)
    self.assertEqual(tbl[1, 0], 12)

  def test11_filtering_segmentation(self):
    # """ filtering with segmentation """
    self.details.tableName = 'ferro_noquant'
    with open(os.path.join(self.baseDir, 'ferromag_filt_10_3.pkl'), 'r') as pklTF:
      buf = pklTF.read().replace('\r\n', '\n').encode('utf-8')
      pklTF.close()
    with io.BytesIO(buf) as pklF:
      compos = pickle.load(pklF)
    tgt = 10
    self.assertEqual(len(compos), tgt)
    self.details.doHoldout = 1
    self.details.filterVal = 1
    self.details.filterFrac = .33

    nGood, misCount, nSkipped, avgGood, avgBad, avgSkip, tbl = ScreenComposite.ScreenFromDetails(
      compos, self.details)

    self.assertEqual(nGood, 37)
    self.assertEqual(misCount, 6)
    self.assertEqual(nSkipped, 0)
    self.assertAlmostEqual(avgGood, .95946, 4)
    self.assertAlmostEqual(avgBad, .85, 4)
    self.assertAlmostEqual(avgSkip, 0, 4)
    self.assertEqual(tbl[0, 0], 14)
    self.assertEqual(tbl[1, 1], 23)
    self.assertEqual(tbl[0, 1], 1)
    self.assertEqual(tbl[1, 0], 5)

  def test12_naiveBayes_composite(self):
    # """ test the naive bayes composite"""
    self.details.tableName = 'ferro_noquant'
    with open(os.path.join(self.baseDir, 'ferromag_NaiveBayes.pkl'), 'r') as pklTF:
      buf = pklTF.read().replace('\r\n', '\n').encode('utf-8')
      pklTF.close()
    with io.BytesIO(buf) as pklF:
      compos = pickle.load(pklF)
    tgt = 10
    self.assertEqual(len(compos), tgt)
    self.details.doHoldout = 1
    nGood, misCount, nSkipped, avgGood, avgBad, avgSkip, tbl = ScreenComposite.ScreenFromDetails(
      compos, self.details)
    self.assertEqual(nGood, 25)
    self.assertEqual(misCount, 6)
    self.assertEqual(nSkipped, 0)
    self.assertAlmostEqual(avgGood, 0.9800, 4)
    self.assertAlmostEqual(avgBad, 0.86667, 4)
    self.assertAlmostEqual(avgSkip, 0, 4)
    self.assertEqual(tbl[0, 0], 9)
    self.assertEqual(tbl[0, 1], 6)
    self.assertEqual(tbl[1, 0], 0)
    self.assertEqual(tbl[1, 1], 16)


if __name__ == '__main__':  # pragma: nocover
  unittest.main()



tests = [
  ("python", "UnitTestSDMolSupplier.py", {}),
  ("python", "UnitTestSmilesMolSupplier.py", {}),
  ("python", "UnitTestDbMolSupplier.py", {}),
]

longTests = []
if __name__ == '__main__':
  import sys

  from rdkit import TestRunner
  failed, tests = TestRunner.RunScript('test_list.py', 0, 1)
  sys.exit(len(failed))



# Automatically adapted for numpy.oldnumeric Jun 27, 2008 by -c

import os
import sys
import tempfile
#
#  $Id: testDepictor.py 2112 2012-07-02 09:47:45Z glandrum $
#
# pylint:disable=E1101,C0111,C0103,R0904
import unittest

import numpy as np

from rdkit import Chem, Geometry, RDConfig
from rdkit.Chem import rdDepictor, rdMolAlign, rdMolTransforms
from rdkit.Chem.ChemUtils import AlignDepict


def feq(v1, v2, tol2=1e-4):
  return abs(v1 - v2) <= tol2


def ptEq(pt1, pt2, tol=1e-4):
  return feq(pt1.x, pt2.x, tol) and feq(pt1.y, pt2.y, tol) and feq(pt1.z, pt2.z, tol)


def getDistMat(mol):
  conf = mol.GetConformer()
  nat = mol.GetNumAtoms()
  nl = nat * (nat - 1) // 2
  res = np.zeros(nl, float)

  for i in range(1, nat):
    pi = conf.GetAtomPosition(i)
    idx = i * (i - 1) // 2
    for j in range(i):
      pj = conf.GetAtomPosition(j)
      pj -= pi
      res[idx + j] = pj.Length()

  return res


def compareCoords(m, molFile):
  mo = Chem.MolFromMolFile(molFile)
  co = mo.GetConformer()

  ci = m.GetConformer()
  nat = m.GetNumAtoms()
  if (nat != mo.GetNumAtoms()):
    return 0

  for i in range(nat):
    pos = ci.GetAtomPosition(i)
    opos = co.GetAtomPosition(i)
    if not ptEq(pos, opos):
      print(Chem.MolToMolBlock(m))
      print(Chem.MolToMolBlock(mo))
      return 0
  return 1


def compareWithOld(smilesFile, sdFile):
  smiSup = Chem.SmilesMolSupplier(smilesFile, ",", 0, -1)
  sdsup = Chem.SDMolSupplier(sdFile)
  im = 0
  for mol in smiSup:
    omol = sdsup[im]
    rdDepictor.Compute2DCoords(mol, canonOrient=False)
    conf = mol.GetConformer()
    oconf = omol.GetConformer()
    nat = mol.GetNumAtoms()
    for i in range(nat):
      pos = conf.GetAtomPosition(i)
      opos = oconf.GetAtomPosition(i)
      if not ptEq(pos, opos):
        print(Chem.MolToMolBlock(omol), file=sys.stderr)
        print('> <Failed>\n%d\n' % i, file=sys.stderr)
        print("$$$$", file=sys.stderr)
        print(Chem.MolToMolBlock(mol), file=sys.stderr)
        print('> <Failed>\n%d\n' % i, file=sys.stderr)
        print("$$$$", file=sys.stderr)
        return 0
    im += 1
  return 1


def stereoCompare(smilesFile):
  smiSup = Chem.SmilesMolSupplier(smilesFile, ",", 0, -1)
  for mol in smiSup:
    rdDepictor.Compute2DCoords(mol, canonOrient=False)
    mb = Chem.MolToMolBlock(mol)
    nmol = Chem.MolFromMolBlock(mb)
    matches = nmol.GetSubstructMatches(mol, False)
    dbnds = [
      x for x in mol.GetBonds()
      if (x.GetBondType() == Chem.BondType.DOUBLE and x.GetStereo() > Chem.BondStereo.STEREOANY)
    ]
    ok = True
    for match in matches:
      for bnd in dbnds:
        obnd = nmol.GetBondBetweenAtoms(match[bnd.GetBeginAtomIdx()], match[bnd.GetEndAtomIdx()])
        assert (obnd.GetBondType() == Chem.BondType.DOUBLE)
      if ok:
        break
    if not ok:
      print(Chem.MolToMolBlock(mol), file=sys.stderr)
      print("$$$$", file=sys.stderr)
      return 0
  return 1


class TestCase(unittest.TestCase):

  def _test0First200(self):
    # this test is disabled because it's not particularly useful and
    # causes problems every time anything changes.
    fileN = os.path.join(RDConfig.RDBaseDir, 'Code', 'GraphMol', 'Depictor', 'test_data',
                         'first_200.tpsa.csv')
    #smiSup = Chem.SmilesMolSupplier(fileN, ",", 0, -1)

    ofile = os.path.join(RDConfig.RDBaseDir, 'Code', 'GraphMol', 'Depictor', 'test_data',
                         'first_200.python.sdf')
    self.assertTrue(compareWithOld(fileN, ofile))

  def test1CisTrans(self):
    fileN = os.path.join(RDConfig.RDBaseDir, 'Code', 'GraphMol', 'Depictor', 'test_data',
                         "cis_trans_cases.csv")
    self.assertTrue(stereoCompare(fileN))

  def test2Coords(self):
    m1 = Chem.MolFromSmiles('C1CCC1CC')
    coordMap = {
      0: Geometry.Point2D(0, 0),
      1: Geometry.Point2D(1.5, 0),
      2: Geometry.Point2D(1.5, 1.5),
      3: Geometry.Point2D(0, 1.5)
    }
    rdDepictor.Compute2DCoords(m1, coordMap=coordMap)
    conf = m1.GetConformer(0)
    for i in range(4):
      self.assertTrue(
        ptEq(conf.GetAtomPosition(i), Geometry.Point3D(coordMap[i].x, coordMap[i].y, 0.0)))

    m1 = Chem.MolFromSmiles('CCC')
    try:
      rdDepictor.Compute2DCoords(m1, coordMap=coordMap)
      ok = 0
    except ValueError:
      ok = 1
    self.assertTrue(ok)

  def test3IssueSF1526844(self):
    t = Chem.MolFromSmiles('c1nc(N)ccc1')
    rdDepictor.Compute2DCoords(t, canonOrient=False)

    m2 = Chem.MolFromSmiles('c1nc(NC=O)ccc1')
    AlignDepict.AlignDepict(m2, t)
    expected = [
      Geometry.Point3D(1.5, 0.0, 0.0),
      Geometry.Point3D(0.75, -1.299, 0.0),
      Geometry.Point3D(-0.75, -1.299, 0.0),
      Geometry.Point3D(-1.5, -2.5981, 0.0),
      Geometry.Point3D(-3.0, -2.5981, 0.0),
      Geometry.Point3D(-3.75, -3.8971, 0.0),
      Geometry.Point3D(-1.5, 0.0, 0.0),
      Geometry.Point3D(-0.75, 1.2990, 0.0),
      Geometry.Point3D(0.75, 1.2990, 0.0)
    ]

    nat = m2.GetNumAtoms()
    conf = m2.GetConformer()
    for i in range(nat):
      pos = conf.GetAtomPosition(i)
      self.assertTrue(ptEq(pos, expected[i], 0.001))

  def test4SamplingSpread(self):
    mol = Chem.MolFromMolFile(
      os.path.join(RDConfig.RDBaseDir, 'Code/GraphMol/Depictor', 'test_data/7UPJ_xtal.mol'))

    # default mode
    rdDepictor.Compute2DCoords(mol, canonOrient=False)
    self.assertTrue(
      compareCoords(
        mol, os.path.join(RDConfig.RDBaseDir, 'Code/GraphMol/Depictor',
                          'test_data/7UPJ_default.mol')))

    # spread the structure as much as possible by sampling
    rdDepictor.Compute2DCoords(mol, canonOrient=False, nFlipsPerSample=3, nSample=100,
                               sampleSeed=100, permuteDeg4Nodes=1)
    self.assertTrue(
      compareCoords(
        mol, os.path.join(RDConfig.RDBaseDir, 'Code/GraphMol/Depictor',
                          'test_data/7UPJ_spread.mol')))

  def test5SamplingMimic3D(self):
    mol = Chem.MolFromMolFile(
      os.path.join(RDConfig.RDBaseDir, 'Code/GraphMol/Depictor', 'test_data/7UPJ_xtal.mol'))
    dmat3D = getDistMat(mol)

    # now mimic the coordinate with a very small weight
    rdDepictor.Compute2DCoordsMimicDistmat(mol, dmat3D, weightDistMat=0.001)
    self.assertTrue(
      compareCoords(
        mol,
        os.path.join(RDConfig.RDBaseDir, 'Code/GraphMol/Depictor', 'test_data/7UPJ_mimic3D_1.mol')))

    # now mimic the coordinate with a very small weight
    rdDepictor.Compute2DCoordsMimicDistmat(mol, dmat3D, weightDistMat=0.003)
    self.assertTrue(
      compareCoords(
        mol,
        os.path.join(RDConfig.RDBaseDir, 'Code/GraphMol/Depictor', 'test_data/7UPJ_mimic3D_2.mol')))

    #mb = Chem.MolToMolBlock(mol)
    #ofile = open('../test_data/7UPJ_mimic3D_2.mol', 'w')
    # ofile.write(mb)
    # ofile.close()

  def test6ChangeBondLength(self):
    m = Chem.MolFromSmiles('CC')
    rdDepictor.Compute2DCoords(m)
    conf = m.GetConformer()
    self.assertAlmostEqual(conf.GetAtomPosition(0).x, -0.750, 3)
    self.assertAlmostEqual(conf.GetAtomPosition(1).x, 0.750, 3)
    rdDepictor.Compute2DCoords(m, bondLength=1.0)
    conf = m.GetConformer()
    self.assertAlmostEqual(conf.GetAtomPosition(0).x, -0.500, 3)
    self.assertAlmostEqual(conf.GetAtomPosition(1).x, 0.500, 3)
    rdDepictor.Compute2DCoords(m)
    conf = m.GetConformer()
    self.assertAlmostEqual(conf.GetAtomPosition(0).x, -0.750, 3)
    self.assertAlmostEqual(conf.GetAtomPosition(1).x, 0.750, 3)

  def testConstrainedCoords(self):
    templ = Chem.MolFromSmiles('c1nccc2n1ccc2')
    rdDepictor.Compute2DCoords(templ)
    m1 = Chem.MolFromSmiles('c1cccc2ncn3cccc3c21')
    rdDepictor.GenerateDepictionMatching2DStructure(m1, templ)
    m2 = Chem.MolFromSmiles('c1cc(Cl)cc2ncn3cccc3c21')
    rdDepictor.Compute2DCoords(m2)
    refPatt1 = Chem.MolFromSmarts('*1****2*1***2')
    rdDepictor.GenerateDepictionMatching2DStructure(m2, templ, -1, refPatt1)
    fileN = os.path.join(RDConfig.RDBaseDir, 'Code', 'GraphMol', 'Depictor', 'test_data',
                         '1XP0_ligand.sdf')

    xp0_lig = Chem.MolFromMolFile(fileN)
    xp0_lig_2d = Chem.Mol(xp0_lig)
    rdDepictor.GenerateDepictionMatching3DStructure(xp0_lig_2d, xp0_lig)
    xp0_ref = Chem.MolFromSmarts('[#6]1~[#7][#6]~[#6]2[#6](=[#8])[#7]~[#6](c3ccccc3)[#7][#7]12')
    rdDepictor.GenerateDepictionMatching3DStructure(xp0_lig_2d, xp0_lig, -1, xp0_ref)

  def testGenerate2DDepictionRefPatternAtomMap(self):
    indazoleMolblock = """
     RDKit          2D

  9 10  0  0  0  0  0  0  0  0999 V2000
   -6.0878    2.4335    0.0000 C   0  0  0  0  0  0  0  0  0  0  0  0
   -7.3867    1.6835    0.0000 C   0  0  0  0  0  0  0  0  0  0  0  0
   -7.3867    0.1833    0.0000 C   0  0  0  0  0  0  0  0  0  0  0  0
   -6.0878   -0.5666    0.0000 C   0  0  0  0  0  0  0  0  0  0  0  0
   -4.7887    0.1833    0.0000 C   0  0  0  0  0  0  0  0  0  0  0  0
   -4.7887    1.6835    0.0000 C   0  0  0  0  0  0  0  0  0  0  0  0
   -3.4897   -0.5664    0.0000 N   0  0  0  0  0  0  0  0  0  0  0  0
   -2.1906    1.6833    0.0000 C   0  0  0  0  0  0  0  0  0  0  0  0
   -2.1906    0.1835    0.0000 N   0  0  0  0  0  0  0  0  0  0  0  0
  1  2  2  0
  2  3  1  0
  3  4  2  0
  4  5  1  0
  5  6  2  0
  6  1  1  0
  8  9  2  0
  6  8  1  0
  7  9  1  0
  7  5  1  0
M  END"""
    indazoleRef = Chem.MolFromMolBlock(indazoleMolblock)
    cycloheptylPyrazole = Chem.MolFromSmiles("c1cc(C2CCCCCC2)[nH]n1")

    # test using refPattern
    refPatt = Chem.MolFromSmarts("a1aan[nH]1")
    rdDepictor.GenerateDepictionMatching2DStructure(cycloheptylPyrazole, indazoleRef,
                                                    refPatt=refPatt)
    self.assertEqual(cycloheptylPyrazole.GetNumConformers(), 1)
    molMatchVect = cycloheptylPyrazole.GetSubstructMatch(refPatt)
    self.assertEqual(len(molMatchVect), refPatt.GetNumAtoms())
    refMatchVect = indazoleRef.GetSubstructMatch(refPatt)
    self.assertEqual(len(refMatchVect), refPatt.GetNumAtoms())
    atomMap = tuple(zip(refMatchVect, molMatchVect))
    msd = 0.0
    for refIdx, molIdx in atomMap:
      msd += (indazoleRef.GetConformer().GetAtomPosition(refIdx) -
              cycloheptylPyrazole.GetConformer().GetAtomPosition(molIdx)).LengthSq()
    msd /= len(molMatchVect)
    self.assertAlmostEqual(msd, 0.0)
    # try with a pattern larger than the reference molecule
    hugePatt = Chem.MolFromSmarts("CCCCCCCCCCCCCCCCCCCCCCCCCCC")
    with self.assertRaises(ValueError):
      rdDepictor.GenerateDepictionMatching2DStructure(cycloheptylPyrazole, indazoleRef,
                                                      refPatt=hugePatt)

    # try with an out of range confId
    with self.assertRaises(ValueError):
      rdDepictor.GenerateDepictionMatching2DStructure(cycloheptylPyrazole, indazoleRef, confId=1,
                                                      refPatt=refPatt)

    # test using atomMap directly
    cycloheptylPyrazole.RemoveAllConformers()
    rdDepictor.GenerateDepictionMatching2DStructure(cycloheptylPyrazole, indazoleRef,
                                                    atomMap=atomMap)
    self.assertEqual(cycloheptylPyrazole.GetNumConformers(), 1)
    msd = 0.0
    for refIdx, molIdx in atomMap:
      msd += (indazoleRef.GetConformer().GetAtomPosition(refIdx) -
              cycloheptylPyrazole.GetConformer().GetAtomPosition(molIdx)).LengthSq()
    msd /= len(atomMap)
    self.assertAlmostEqual(msd, 0.0)

    # try with an atomMap larger than the reference molecule
    atomMapHuge = list(atomMap) + [(0, 0) for i in range(indazoleRef.GetNumAtoms())]
    with self.assertRaises(ValueError):
      rdDepictor.GenerateDepictionMatching2DStructure(cycloheptylPyrazole, indazoleRef,
                                                      atomMap=atomMapHuge)

    # try with an atomMap with out of range indices
    atomMapOutOfRange = list(atomMap) + [(100, 100)]
    with self.assertRaises(ValueError):
      rdDepictor.GenerateDepictionMatching2DStructure(cycloheptylPyrazole, indazoleRef,
                                                      atomMap=atomMapOutOfRange)

    # try with an out of range confId
    with self.assertRaises(ValueError):
      rdDepictor.GenerateDepictionMatching2DStructure(cycloheptylPyrazole, indazoleRef,
                                                      atomMap=atomMap, confId=1)

  def testGenerate2DDepictionAllowRGroups(self):
    templateMolblock = """
     RDKit          2D

  9  9  0  0  0  0  0  0  0  0999 V2000
   -0.8929    1.0942    0.0000 C   0  0  0  0  0  0  0  0  0  0  0  0
   -2.1919    0.3442    0.0000 C   0  0  0  0  0  0  0  0  0  0  0  0
   -2.1919   -1.1558    0.0000 C   0  0  0  0  0  0  0  0  0  0  0  0
   -0.8929   -1.9059    0.0000 C   0  0  0  0  0  0  0  0  0  0  0  0
    0.4060   -1.1558    0.0000 C   0  0  0  0  0  0  0  0  0  0  0  0
    0.4060    0.3442    0.0000 C   0  0  0  0  0  0  0  0  0  0  0  0
   -3.4910    1.0942    0.0000 R1  0  0  0  0  0  0  0  0  0  0  0  0
    1.7051    1.0942    0.0000 R2  0  0  0  0  0  0  0  0  0  0  0  0
   -3.4910   -1.9059    0.0000 R3  0  0  0  0  0  0  0  0  0  0  0  0
  1  2  2  0
  2  3  1  0
  3  4  2  0
  4  5  1  0
  5  6  2  0
  6  1  1  0
  6  8  1  0
  3  9  1  0
  2  7  1  0
M  RGP  3   7   1   8   2   9   3
M  END"""
    templateRef = Chem.MolFromMolBlock(templateMolblock)
    orthoMeta = Chem.MolFromSmiles("c1ccc(-c2ccc(-c3ccccc3)c(-c3ccccc3)c2)cc1")
    ortho = Chem.MolFromSmiles("c1ccc(-c2ccccc2-c2ccccc2)cc1")
    meta = Chem.MolFromSmiles("c1ccc(-c2cccc(-c3ccccc3)c2)cc1")
    para = Chem.MolFromSmiles("c1ccc(-c2ccc(-c3ccccc3)cc2)cc1")
    biphenyl = Chem.MolFromSmiles("c1ccccc1-c1ccccc1")
    phenyl = Chem.MolFromSmiles("c1ccccc1")

    atomMap = rdDepictor.GenerateDepictionMatching2DStructure(orthoMeta, templateRef)
    self.assertEqual(orthoMeta.GetNumConformers(), 1)


    # test original usage pattern
    for mol in (ortho, meta, biphenyl, phenyl):
      # fails as does not match template
      with self.assertRaises(ValueError):
        rdDepictor.GenerateDepictionMatching2DStructure(mol, templateRef)

      # succeeds with allowRGroups=true
      atomMap = rdDepictor.GenerateDepictionMatching2DStructure(mol, templateRef, allowRGroups=True)
      self.assertEqual(mol.GetNumConformers(), 1)
      msd = 0.0
      for refIdx, molIdx in atomMap:
        msd += (templateRef.GetConformer().GetAtomPosition(refIdx) -
                mol.GetConformer().GetAtomPosition(molIdx)).LengthSq()
      msd /= len(atomMap)
      self.assertAlmostEqual(msd, 0.0)

    for alignOnly in (True, False):
      p = rdDepictor.ConstrainedDepictionParams()
      p.allowRGroups = True
      p.alignOnly = alignOnly
      for mol in (ortho, meta, para, biphenyl, phenyl):
        # fails as does not match template
        with self.assertRaises(ValueError):
          rdDepictor.GenerateDepictionMatching2DStructure(mol, templateRef)

        # succeeds with allowRGroups=true
        atomMap = rdDepictor.GenerateDepictionMatching2DStructure(mol, templateRef, params=p)
        self.assertGreater(len(atomMap), 0)
        self.assertEqual(mol.GetNumConformers(), 1)
        msd = 0.0
        for refIdx, molIdx in atomMap:
          msd += (templateRef.GetConformer().GetAtomPosition(refIdx) -
                  mol.GetConformer().GetAtomPosition(molIdx)).LengthSq()
        msd /= len(atomMap)
        self.assertAlmostEqual(msd, 0.0)
        if not p.alignOnly:
          self.assertEqual(atomMap, rdDepictor.GenerateDepictionMatching2DStructure(mol, templateRef, allowRGroups=True))

      # test that using a refPattern with R groups and a reference without works
      pyridineRef = Chem.MolFromMolBlock("""
     RDKit          2D

  8  8  0  0  0  0  0  0  0  0999 V2000
    0.0000    1.5469    0.0000 C   0  0  0  0  0  0  0  0  0  0  0  0
   -1.3395    0.7734    0.0000 C   0  0  0  0  0  0  0  0  0  0  0  0
   -1.3395   -0.7732    0.0000 C   0  0  0  0  0  0  0  0  0  0  0  0
    0.0000   -1.5469    0.0000 C   0  0  0  0  0  0  0  0  0  0  0  0
    1.3395   -0.7732    0.0000 C   0  0  0  0  0  0  0  0  0  0  0  0
    1.3395    0.7734    0.0000 N   0  0  0  0  0  0  0  0  0  0  0  0
    0.0000    3.0938    0.0000 C   0  0  0  0  0  0  0  0  0  0  0  0
    0.0000   -3.0938    0.0000 C   0  0  0  0  0  0  0  0  0  0  0  0
  1  2  2  0
  2  3  1  0
  3  4  2  0
  4  5  1  0
  5  6  2  0
  6  1  1  0
  1  7  1  0
  4  8  1  0
M  END""")
      genericRefPatternWithRGroups = Chem.MolFromSmarts("[*:3]a1a([*:1])aa([*:2])aa1")

      for numExpectedMatches, mol in (
        (8, orthoMeta), (7, ortho), (7, meta), (8, para), (7, biphenyl), (6, phenyl)
      ):
        atomMap = rdDepictor.GenerateDepictionMatching2DStructure(
          mol, pyridineRef, -1, genericRefPatternWithRGroups, p)
        self.assertEqual(len(atomMap), numExpectedMatches)
        self.assertEqual(mol.GetNumConformers(), 1)
        msd = 0.0
        for refIdx, molIdx in atomMap:
          msd += (pyridineRef.GetConformer().GetAtomPosition(refIdx) -
                  mol.GetConformer().GetAtomPosition(molIdx)).LengthSq()
        msd /= len(atomMap)
        self.assertLess(msd, 5.e-3 if alignOnly else 1.e-4)

  def testNormalizeStraighten(self):
    noradrenalineMJ = Chem.MolFromMolBlock("""
  MJ201100                      

 12 12  0  0  1  0  0  0  0  0999 V2000
    2.2687    1.0716    0.0000 N   0  0  0  0  0  0  0  0  0  0  0  0
    1.4437    1.0716    0.0000 C   0  0  0  0  0  0  0  0  0  0  0  0
    1.0312    0.3572    0.0000 C   0  0  0  0  0  0  0  0  0  0  0  0
    1.4437   -0.3572    0.0000 O   0  0  0  0  0  0  0  0  0  0  0  0
    0.2062    0.3572    0.0000 C   0  0  0  0  0  0  0  0  0  0  0  0
   -0.2062   -0.3572    0.0000 C   0  0  0  0  0  0  0  0  0  0  0  0
   -1.0312   -0.3572    0.0000 C   0  0  0  0  0  0  0  0  0  0  0  0
   -1.4437   -1.0716    0.0000 O   0  0  0  0  0  0  0  0  0  0  0  0
   -1.4437    0.3572    0.0000 C   0  0  0  0  0  0  0  0  0  0  0  0
   -2.2687    0.3572    0.0000 O   0  0  0  0  0  0  0  0  0  0  0  0
   -1.0312    1.0716    0.0000 C   0  0  0  0  0  0  0  0  0  0  0  0
   -0.2062    1.0716    0.0000 C   0  0  0  0  0  0  0  0  0  0  0  0
  1  2  1  0  0  0  0
  3  2  1  0  0  0  0
  3  4  1  6  0  0  0
  3  5  1  0  0  0  0
  5  6  2  0  0  0  0
  6  7  1  0  0  0  0
  7  8  1  0  0  0  0
  7  9  2  0  0  0  0
  9 10  1  0  0  0  0
  9 11  1  0  0  0  0
 11 12  2  0  0  0  0
  5 12  1  0  0  0  0
M  END)""")

    noradrenalineMJCopy = Chem.Mol(noradrenalineMJ)
    conformer0 = noradrenalineMJCopy.GetConformer(0)
    conformer1 = Chem.Conformer(conformer0)
    noradrenalineMJCopy.AddConformer(conformer1, True)
    conformer1 = noradrenalineMJCopy.GetConformer(1)
    self.assertLess(rdMolAlign.CalcRMS(noradrenalineMJ, noradrenalineMJCopy, 0, 0), 1.e-5)
    self.assertLess(rdMolAlign.CalcRMS(noradrenalineMJ, noradrenalineMJCopy, 0, 1), 1.e-5)
    scalingFactor = rdDepictor.NormalizeDepiction(noradrenalineMJCopy, 1)
    self.assertLess(rdMolAlign.CalcRMS(noradrenalineMJ, noradrenalineMJCopy, 0, 0), 1.e-5)
    self.assertGreater(rdMolAlign.CalcRMS(noradrenalineMJ, noradrenalineMJCopy, 0, 1), 1.e-5)
    self.assertAlmostEqual(scalingFactor, 1.875, 3)
    conformer2 = Chem.Conformer(conformer1)
    noradrenalineMJCopy.AddConformer(conformer2, True)
    conformer2 = noradrenalineMJCopy.GetConformer(2)
    bond10_11Conf0 = conformer0.GetAtomPosition(11) - conformer0.GetAtomPosition(10)
    self.assertAlmostEqual(bond10_11Conf0.x, 0.825, 3)
    self.assertAlmostEqual(bond10_11Conf0.y, 0., 3)
    bond10_11Conf1 = conformer1.GetAtomPosition(11) - conformer1.GetAtomPosition(10)
    self.assertAlmostEqual(bond10_11Conf1.x, 1.513, 3)
    self.assertAlmostEqual(bond10_11Conf1.y, -0.321, 3)
    rdDepictor.StraightenDepiction(noradrenalineMJCopy, 1)
    bond10_11Conf1 = conformer1.GetAtomPosition(11) - conformer1.GetAtomPosition(10)
    self.assertAlmostEqual(bond10_11Conf1.x, 1.340, 3)
    self.assertAlmostEqual(bond10_11Conf1.y, -0.773, 3)
    bond4_11Conf1 = conformer1.GetAtomPosition(11) - conformer1.GetAtomPosition(4)
    self.assertAlmostEqual(bond4_11Conf1.x, 0., 3)
    self.assertAlmostEqual(bond4_11Conf1.y, 1.547, 3)
    rdDepictor.StraightenDepiction(noradrenalineMJCopy, 2, True)
    bond10_11Conf2 = conformer2.GetAtomPosition(11) - conformer2.GetAtomPosition(10)
    self.assertAlmostEqual(bond10_11Conf2.x, 1.547, 3)
    self.assertAlmostEqual(bond10_11Conf2.y, 0.0, 3)
    bond4_11Conf2 = conformer2.GetAtomPosition(11) - conformer2.GetAtomPosition(4)
    self.assertAlmostEqual(bond4_11Conf2.x, -0.773, 3)
    self.assertAlmostEqual(bond4_11Conf2.y, 1.339, 3)

    noradrenalineMJCopy = Chem.Mol(noradrenalineMJ)
    conformer0 = noradrenalineMJCopy.GetConformer(0)
    conformer1 = Chem.Conformer(conformer0)
    noradrenalineMJCopy.AddConformer(conformer1, True)
    conformer1 = noradrenalineMJCopy.GetConformer(1)
    scalingFactor = rdDepictor.NormalizeDepiction(noradrenalineMJCopy, 1, -1)
    self.assertLess(rdMolAlign.CalcRMS(noradrenalineMJ, noradrenalineMJCopy, 0, 0), 1.e-5)
    self.assertGreater(rdMolAlign.CalcRMS(noradrenalineMJ, noradrenalineMJCopy, 0, 1), 1.e-5)
    self.assertAlmostEqual(scalingFactor, 1.875, 3)
    conformer2 = Chem.Conformer(conformer1)
    noradrenalineMJCopy.AddConformer(conformer2, True)
    conformer2 = noradrenalineMJCopy.GetConformer(2)
    bond10_11Conf0 = conformer0.GetAtomPosition(11) - conformer0.GetAtomPosition(10)
    self.assertAlmostEqual(bond10_11Conf0.x, 0.825, 3)
    self.assertAlmostEqual(bond10_11Conf0.y, 0., 3)
    bond10_11Conf1 = conformer1.GetAtomPosition(11) - conformer1.GetAtomPosition(10)
    self.assertAlmostEqual(bond10_11Conf1.x, 0.321, 3)
    self.assertAlmostEqual(bond10_11Conf1.y, 1.513, 3)
    rdDepictor.StraightenDepiction(noradrenalineMJCopy, 1)
    bond10_11Conf1 = conformer1.GetAtomPosition(11) - conformer1.GetAtomPosition(10)
    self.assertAlmostEqual(bond10_11Conf1.x, 0.0, 3)
    self.assertAlmostEqual(bond10_11Conf1.y, 1.547, 3)
    rdDepictor.StraightenDepiction(noradrenalineMJCopy, 2, True)
    bond10_11Conf2 = conformer2.GetAtomPosition(11) - conformer2.GetAtomPosition(10)
    self.assertAlmostEqual(bond10_11Conf2.x, bond10_11Conf1.x, 3)
    self.assertAlmostEqual(bond10_11Conf2.y, bond10_11Conf1.y, 3)

    noradrenalineMJCopy = Chem.Mol(noradrenalineMJ)
    conformer0 = noradrenalineMJCopy.GetConformer(0)
    conformer1 = Chem.Conformer(conformer0)
    noradrenalineMJCopy.AddConformer(conformer1, True)
    conformer1 = noradrenalineMJCopy.GetConformer(1)
    scalingFactor = rdDepictor.NormalizeDepiction(noradrenalineMJCopy, 1, 0, 3.0)
    self.assertLess(rdMolAlign.CalcRMS(noradrenalineMJ, noradrenalineMJCopy, 0, 0), 1.e-5)
    self.assertGreater(rdMolAlign.CalcRMS(noradrenalineMJ, noradrenalineMJCopy, 0, 1), 1.e-5)
    self.assertAlmostEqual(scalingFactor, 3.0, 3)
    conformer2 = Chem.Conformer(conformer1)
    noradrenalineMJCopy.AddConformer(conformer2, True)
    conformer2 = noradrenalineMJCopy.GetConformer(2)
    conformer3 = Chem.Conformer(conformer1)
    noradrenalineMJCopy.AddConformer(conformer3, True)
    conformer3 = noradrenalineMJCopy.GetConformer(3)
    bond10_11Conf0 = conformer0.GetAtomPosition(11) - conformer0.GetAtomPosition(10)
    self.assertAlmostEqual(bond10_11Conf0.x, 0.825, 3)
    self.assertAlmostEqual(bond10_11Conf0.y, 0., 3)
    bond10_11Conf1 = conformer1.GetAtomPosition(11) - conformer1.GetAtomPosition(10)
    self.assertAlmostEqual(bond10_11Conf1.x, 2.475, 3)
    self.assertAlmostEqual(bond10_11Conf1.y, 0., 3)
    rdDepictor.StraightenDepiction(noradrenalineMJCopy, 1)
    bond10_11Conf1 = conformer1.GetAtomPosition(11) - conformer1.GetAtomPosition(10)
    self.assertAlmostEqual(bond10_11Conf1.x, 2.143, 3)
    self.assertAlmostEqual(bond10_11Conf1.y, -1.237, 3)
    bond4_11Conf1 = conformer1.GetAtomPosition(11) - conformer1.GetAtomPosition(4)
    self.assertAlmostEqual(bond4_11Conf1.x, 0., 3)
    self.assertAlmostEqual(bond4_11Conf1.y, 2.475, 3)
    rdDepictor.StraightenDepiction(noradrenalineMJCopy, 2, True)
    bond10_11Conf2 = conformer2.GetAtomPosition(11) - conformer2.GetAtomPosition(10)
    bond10_11Conf3 = conformer3.GetAtomPosition(11) - conformer3.GetAtomPosition(10)
    self.assertAlmostEqual(bond10_11Conf2.x, bond10_11Conf3.x, 3)
    self.assertAlmostEqual(bond10_11Conf2.y, bond10_11Conf3.y, 3)
    bond4_11Conf2 = conformer2.GetAtomPosition(11) - conformer2.GetAtomPosition(4)
    bond4_11Conf3 = conformer3.GetAtomPosition(11) - conformer3.GetAtomPosition(4)
    self.assertAlmostEqual(bond4_11Conf2.x, bond4_11Conf3.x, 3)
    self.assertAlmostEqual(bond4_11Conf2.y, bond4_11Conf3.y, 3)

  @unittest.skipIf(not rdDepictor.IsCoordGenSupportAvailable(), "CoordGen not available, skipping")
  def testUsingCoordGenCtxtMgr(self):
    default_status = rdDepictor.GetPreferCoordGen()

    # This is the default; we shouldn't have changed it
    self.assertEqual(default_status, False)

    with rdDepictor.UsingCoordGen(True):
      current_status = rdDepictor.GetPreferCoordGen()
      self.assertEqual(current_status, True)

    current_status = rdDepictor.GetPreferCoordGen()
    self.assertEqual(current_status, False)

    rdDepictor.SetPreferCoordGen(True)

    with rdDepictor.UsingCoordGen(False):
      current_status = rdDepictor.GetPreferCoordGen()
      self.assertEqual(current_status, False)

    current_status = rdDepictor.GetPreferCoordGen()
    self.assertEqual(current_status, True)

    rdDepictor.SetPreferCoordGen(default_status)

  def molMatchesTemplate(self, mol, template):
    """
    Determines if the shape/layout of the template and mol are the same. It
    is ok if the mol and template are not centered at the same place, or if
    the mol and template have different orientations.
    """
    match = mol.GetSubstructMatch(template)
    if not match or len(match) != template.GetNumAtoms():
      return False

    # get positions of atoms with centroid at origin, it is ok if the
    # template or mol is not centered
    template_match_positions = [
      mol.GetConformer().GetPositions()[mol_at_idx] for mol_at_idx in match
    ]
    template_match_center = sum(template_match_positions) / len(template_match_positions)
    mol_positions = [p - template_match_center for p in mol.GetConformer().GetPositions()]

    template_center = sum(template.GetConformer().GetPositions()) / template.GetNumAtoms()
    template_positions = [p - template_center for p in template.GetConformer().GetPositions()]

    # the mol may match the template but be slightly rotated about the centroid
    # or reflected across the x or y axis
    rotations = [[], [], [], []]
    for template_idx, idx in enumerate(match):
      v1 = mol_positions[idx]

      # no reflection
      v2 = template_positions[template_idx]
      val = round(np.dot(v1, v2) / (np.linalg.norm(v1) * np.linalg.norm(v2)), 4)
      rotations[0].append(np.arccos(val))

      # reflect across x-axis
      v2[0] = v2[0] * -1
      val = round(np.dot(v1, v2) / (np.linalg.norm(v1) * np.linalg.norm(v2)), 4)
      rotations[1].append(np.arccos(val))

      # reflect across y-axis
      v2[0] = v2[0] * -1
      v2[1] = v2[1] * -1
      val = round(np.dot(v1, v2) / (np.linalg.norm(v1) * np.linalg.norm(v2)), 4)
      rotations[2].append(np.arccos(val))

      # reflect across y-axis and x-acis
      v2[0] = v2[0] * -1
      # v2[1] = v2[1] * -1
      val = round(np.dot(v1, v2) / (np.linalg.norm(v1) * np.linalg.norm(v2)), 4)
      rotations[3].append(np.arccos(val))

    # if all the rotations are similar, then the shape is the same
    return np.any([np.allclose(r, r[0], atol=.05) for r in rotations])

  def assertMolMatchesCoordMap(self, mol, coord_map):
    for aid, expected_position in coord_map.items():
      actual_position = mol.GetConformer().GetAtomPosition(aid)
      self.assertAlmostEqual(actual_position.x, expected_position.x)
      self.assertAlmostEqual(actual_position.y, actual_position.y)

  def testUseMultipleTemplates(self):
    with rdDepictor.UsingCoordGen(False):
      # templates that will be linked together
      template1 = Chem.MolFromSmiles(
        "C1=CCCC2CCCCC2CCCCC2CCCC(CCCCCCC1)C2 |(-0.04,3.43,;-0.04,1.93,;-1.34,1.18,;-2.64,1.93,;-3.94,1.18,;-5.24,1.93,;-6.54,1.18,;-6.54,-0.32,;-5.24,-1.07,;-3.94,-0.32,;-2.64,-1.07,;-2.64,-2.57,;-1.34,-3.32,;-0.04,-2.57,;1.26,-3.32,;1.26,-4.82,;2.56,-5.56,;3.86,-4.82,;3.86,-3.32,;5.16,-2.57,;5.16,-1.07,;3.86,-0.32,;3.86,1.18,;2.56,1.93,;2.56,3.43,;1.26,4.18,;2.56,-2.57,)|"
      )
      template2 = Chem.MolFromSmiles(
        "C1CCC2C(C1)C1CCN2NN1 |(-2.94,-0.77,;-2.94,0.77,;-1.6,1.54,;-0.27,0.77,;-0.27,-0.77,;-1.6,-1.54,;1.06,-1.54,;2.4,-0.77,;2.4,0.77,;1.06,1.54,;1.33,0.51,;1.33,-0.51,)|"
      )
      template3 = Chem.MolFromSmiles(
        "C1C2CC3CC1CC3C2 |(-7.01,3.13,;-7.71,4.35,;-7.01,5.56,;-5.61,5.56,;-4.91,4.35,;-5.61,3.13,;-4.28,3.57,;-4.28,5.13,;-6.34,4.05,)|"
      )

      # example with 2 templates linked together
      two_linked_templates = Chem.MolFromSmiles(
        "NC(CCC1CCC2C(C1)C1CCN2NN1)CC(=O)CCC1=CCCC2CCCCC2CCCCC2CCCC(CCCCCCC1)C2")
      rdDepictor.Compute2DCoords(two_linked_templates, useRingTemplates=False)
      assert not self.molMatchesTemplate(two_linked_templates, template1)
      assert not self.molMatchesTemplate(two_linked_templates, template2)

      rdDepictor.Compute2DCoords(two_linked_templates, useRingTemplates=True)
      assert self.molMatchesTemplate(two_linked_templates, template1)
      assert self.molMatchesTemplate(two_linked_templates, template2)

      # example with 3 templates linked together
      three_linked_templates = Chem.MolFromSmiles(
        "NC(CCC1CCC2C(C1)C1CC(CCC(=O)CC(N)CC~C3C4CC5CC3CC5C4)N2NN1)CC(=O)CCC1=CCCC2CCCCC2CCCCC2CCCC(CCCCCCC1)C2"
      )
      rdDepictor.Compute2DCoords(three_linked_templates, useRingTemplates=False)
      assert not self.molMatchesTemplate(three_linked_templates, template1)
      assert not self.molMatchesTemplate(three_linked_templates, template2)
      assert not self.molMatchesTemplate(three_linked_templates, template3)

      rdDepictor.Compute2DCoords(three_linked_templates, useRingTemplates=True)
      assert self.molMatchesTemplate(three_linked_templates, template1)
      assert self.molMatchesTemplate(three_linked_templates, template2)
      assert self.molMatchesTemplate(three_linked_templates, template3)

  def testUseTemplateAndCoordMap(self):
    with rdDepictor.UsingCoordGen(False):
      template1 = Chem.MolFromSmiles(
        "C1=CCCC2CCCCC2CCCCC2CCCC(CCCCCCC1)C2 |(-0.04,3.43,;-0.04,1.93,;-1.34,1.18,;-2.64,1.93,;-3.94,1.18,;-5.24,1.93,;-6.54,1.18,;-6.54,-0.32,;-5.24,-1.07,;-3.94,-0.32,;-2.64,-1.07,;-2.64,-2.57,;-1.34,-3.32,;-0.04,-2.57,;1.26,-3.32,;1.26,-4.82,;2.56,-5.56,;3.86,-4.82,;3.86,-3.32,;5.16,-2.57,;5.16,-1.07,;3.86,-0.32,;3.86,1.18,;2.56,1.93,;2.56,3.43,;1.26,4.18,;2.56,-2.57,)|"
      )
      template2 = Chem.MolFromSmiles(
        "C1CCC2C(C1)C1CCN2NN1 |(-2.94,-0.77,;-2.94,0.77,;-1.6,1.54,;-0.27,0.77,;-0.27,-0.77,;-1.6,-1.54,;1.06,-1.54,;2.4,-0.77,;2.4,0.77,;1.06,1.54,;1.33,0.51,;1.33,-0.51,)|"
      )
      two_linked_templates = Chem.MolFromSmiles(
        "NC(CCC1CCC2C(C1)C1CCN2NN1)CC(=O)CCC1=CCCC2CCCCC2CCCCC2CCCC(CCCCCCC1)C2")

      # when a coord map doesn't contain any part of a ring system, ring system
      # templates should still be adhered to
      linker_coord_map = {
        16: Geometry.Point2D(1.5, 0),
        17: Geometry.Point2D(1.5, 1.5),
        19: Geometry.Point2D(0, 1.5)
      }
      rdDepictor.Compute2DCoords(two_linked_templates, coordMap=linker_coord_map,
                                  useRingTemplates=True)
      self.assertMolMatchesCoordMap(two_linked_templates, linker_coord_map)
      assert self.molMatchesTemplate(two_linked_templates, template1)
      assert self.molMatchesTemplate(two_linked_templates, template2)

      # when a coord map contains a partial ring system, ring system templates
      # should not be used because they could be distorted by the user-provided
      # templates
      ring_system_coord_map = {
        31: Geometry.Point2D(1.5, 0),
        32: Geometry.Point2D(1.5, 1.5),
        33: Geometry.Point2D(0, 1.5)
      }
      rdDepictor.Compute2DCoords(two_linked_templates, coordMap=ring_system_coord_map,
                                  useRingTemplates=True)
      self.assertMolMatchesCoordMap(two_linked_templates, ring_system_coord_map)
      # atoms 10, 11, and 13 are in this template so the ring template should not be used
      assert not self.molMatchesTemplate(two_linked_templates, template1)
      assert self.molMatchesTemplate(two_linked_templates, template2)

      # when a coord map contains a single atom, even if it is a part of a ring
      # system, ring system templates should be used and the coord map should be
      # followed
      single_atom_coord_map = {10: Geometry.Point2D(0, 0)}
      rdDepictor.Compute2DCoords(two_linked_templates, coordMap=single_atom_coord_map,
                                  useRingTemplates=True)
      self.assertMolMatchesCoordMap(two_linked_templates, single_atom_coord_map)
      assert self.molMatchesTemplate(two_linked_templates, template1)
      assert self.molMatchesTemplate(two_linked_templates, template2)

  def testSetRingSystemTemplates(self):
    with rdDepictor.UsingCoordGen(False):
      mol = Chem.MolFromSmiles("C1CC2CCOC3OC4CCC(C1)C23OO4")
      default_template = Chem.MolFromSmiles(
        "C1CC2CCOC3OC4CCC(C1)C23OO4 |(3.53,-1.22,;3.53,0.3,;2.21,1.06,;2.21,2.59,;0.89,3.35,;-0.43,2.59,;-0.43,1.06,;-1.9,0.65,;-2.47,-0.76,;-1.71,-2.08,;-0.2,-2.29,;0.89,-1.22,;2.21,-1.99,;0.89,0.3,;0.12,-0.83,;-1.19,-1.25,)|"
      )
      user_provided_template = Chem.MolFromSmiles(
        "C1CC2CCOC3OC4CCC(C1)C23OO4 |(-0.5537,-3.1595,;-1.6057,-2.003,;-1.4262,-0.4072,;-2.9804,0.0271,;-3.5191,1.502,;-2.2028,2.3562,;-0.6818,1.8511,;1.0592,1.4391,;2.6123,1.8366,;3.5191,0.5341,;2.6067,-0.7521,;1.0061,-0.773,;0.7888,-2.3546,;-0.0405,0.5251,;0.4049,2.3,;1.7604,3.1594,)|"
      )

      # default templates are loaded automatically
      rdDepictor.Compute2DCoords(mol, useRingTemplates=True)
      assert self.molMatchesTemplate(mol, default_template)

      # set to user-provided template, this will delete default templates
      fpath = os.path.join(RDConfig.RDBaseDir, 'Code', 'GraphMol', 'Depictor', 'test_data',
                            'ring_system_templates.smi')
      rdDepictor.SetRingSystemTemplates(fpath)
      rdDepictor.Compute2DCoords(mol, useRingTemplates=True)
      assert self.molMatchesTemplate(mol, user_provided_template)

      # set back to default ring system templates
      rdDepictor.LoadDefaultRingSystemTemplates()
      rdDepictor.Compute2DCoords(mol, useRingTemplates=True)
      assert self.molMatchesTemplate(mol, default_template)

  def testSetBadRingSystemTemplates(self):
    with tempfile.NamedTemporaryFile(delete=False) as tmp_file:
      tmp_file.write(b"invalidsmiles")
      tmp_file.seek(0)
      with self.assertRaises(ValueError):
        rdDepictor.SetRingSystemTemplates(tmp_file.name)

    with tempfile.NamedTemporaryFile(delete=False) as tmp_file:
      # not a ring system
      tmp_file.write(b"C |(-0.5537,-3.1595)|")
      tmp_file.seek(0)
      with self.assertRaises(ValueError):
        rdDepictor.SetRingSystemTemplates(tmp_file.name)

    with tempfile.NamedTemporaryFile(delete=False) as tmp_file:
      # no coordinates
      tmp_file.write(b"C1CCCCC1")
      tmp_file.seek(0)
      with self.assertRaises(ValueError):
        rdDepictor.SetRingSystemTemplates(tmp_file.name)

    with tempfile.NamedTemporaryFile(delete=False) as tmp_file:
      # bridged ring system
      tmp_file.write(b"c1ccccc1-c1ccccc1")
      tmp_file.seek(0)
      with self.assertRaises(ValueError):
        rdDepictor.SetRingSystemTemplates(tmp_file.name)

    # set back to default ring system templates
    rdDepictor.LoadDefaultRingSystemTemplates()

  def testAddRingSystemTemplates(self):
    with rdDepictor.UsingCoordGen(False):
      # this is what is in the default ring system templates
      mol = Chem.MolFromSmiles("C1CCCN2CCCC(CC1)C2")
      default_template = Chem.MolFromSmiles(
        "C1CCCN2CCCC(CC1)C2 |(2.64,1.53,;3.1,0.11,;2.52,-1.28,;1.17,-1.94,;-0.28,-1.57,;-1.7,-2.03,;-2.82,-1.03,;-2.51,0.44,;-1.08,0.9,;-0.13,2.06,;1.35,2.31,;0.03,-0.1,)|"
      )
      new_template_smi = "C1CCCN2CCCC(CC1)C2 |(3.2585,-0.7733,;3.2585,0.7733,;2.225,1.9568,;0.6347,2.17,;-0.8068,1.4216,;-2.4372,1.3464,;-3.2585,0,;-2.4372,-1.3464,;-0.8068,-1.4216,;0.6347,-2.17,;2.225,-1.9569,;-0.6701,0,)|"
      new_template = Chem.MolFromSmiles(new_template_smi)

      rdDepictor.Compute2DCoords(mol, useRingTemplates=True)
      assert self.molMatchesTemplate(mol, default_template)

      with tempfile.NamedTemporaryFile(delete=False) as tmp_file:
        tmp_file.write(new_template_smi.encode('utf-8'))
        tmp_file.seek(0)
        rdDepictor.AddRingSystemTemplates(tmp_file.name)

      # now the new template should be used
      rdDepictor.Compute2DCoords(mol, useRingTemplates=True)
      assert self.molMatchesTemplate(mol, new_template)

    # set back to default ring system templates
    rdDepictor.LoadDefaultRingSystemTemplates()

  def testGenerateAlignedCoordsAcceptFailure(self):
    template_ref_molblock = """
     RDKit          2D

  9  9  0  0  0  0  0  0  0  0999 V2000
   -0.8929    1.0942    0.0000 C   0  0  0  0  0  0  0  0  0  0  0  0
   -2.1919    0.3442    0.0000 C   0  0  0  0  0  0  0  0  0  0  0  0
   -2.1919   -1.1558    0.0000 C   0  0  0  0  0  0  0  0  0  0  0  0
   -0.8929   -1.9059    0.0000 C   0  0  0  0  0  0  0  0  0  0  0  0
    0.4060   -1.1558    0.0000 C   0  0  0  0  0  0  0  0  0  0  0  0
    0.4060    0.3442    0.0000 C   0  0  0  0  0  0  0  0  0  0  0  0
   -3.4910    1.0942    0.0000 R1  0  0  0  0  0  0  0  0  0  0  0  0
    1.7051    1.0942    0.0000 R2  0  0  0  0  0  0  0  0  0  0  0  0
   -3.4910   -1.9059    0.0000 R3  0  0  0  0  0  0  0  0  0  0  0  0
  1  2  2  0
  2  3  1  0
  3  4  2  0
  4  5  1  0
  5  6  2  0
  6  1  1  0
  6  8  1  0
  3  9  1  0
  2  7  1  0
M  RGP  3   7   1   8   2   9   3
M  END
"""
    template_ref = Chem.MolFromMolBlock(template_ref_molblock)
    mol_molblock = """
     RDKit          2D

  9  9  0  0  0  0  0  0  0  0999 V2000
   -0.8929    1.0942    0.0000 N   0  0  0  0  0  0  0  0  0  0  0  0
   -2.1919    0.3442    0.0000 C   0  0  0  0  0  0  0  0  0  0  0  0
   -2.1919   -1.1558    0.0000 C   0  0  0  0  0  0  0  0  0  0  0  0
   -0.8929   -1.9059    0.0000 C   0  0  0  0  0  0  0  0  0  0  0  0
    0.4060   -1.1558    0.0000 C   0  0  0  0  0  0  0  0  0  0  0  0
    0.4060    0.3442    0.0000 C   0  0  0  0  0  0  0  0  0  0  0  0
   -3.4910    1.0942    0.0000 F   0  0  0  0  0  0  0  0  0  0  0  0
    1.7051    1.0942    0.0000 Cl  0  0  0  0  0  0  0  0  0  0  0  0
   -3.4910   -1.9059    0.0000 Br  0  0  0  0  0  0  0  0  0  0  0  0
  1  2  2  0
  2  3  1  0
  3  4  2  0
  4  5  1  0
  5  6  2  0
  6  1  1  0
  6  8  1  0
  3  9  1  0
  2  7  1  0
M  END
"""
    mol = Chem.MolFromMolBlock(mol_molblock)
    self.assertRaises(ValueError, lambda: rdDepictor.GenerateDepictionMatching2DStructure(mol, template_ref,
      -1, None, False, False, True))
    self.assertEqual(Chem.MolToMolBlock(mol), mol_molblock)
    self.assertEqual(len(rdDepictor.GenerateDepictionMatching2DStructure(mol, template_ref,
      -1, None, True, False, True)), 0)
    self.assertNotEqual(Chem.MolToMolBlock(mol), mol_molblock)
    mol.RemoveAllConformers()
    p = rdDepictor.ConstrainedDepictionParams()
    p.allowRGroups = True
    p.acceptFailure = False
    self.assertEqual(mol.GetNumConformers(), 0)
    self.assertRaises(ValueError, lambda: rdDepictor.GenerateDepictionMatching2DStructure(mol, template_ref,
      -1, None, p))
    self.assertEqual(mol.GetNumConformers(), 0)
    mol.RemoveAllConformers()
    p = rdDepictor.ConstrainedDepictionParams()
    p.allowRGroups = True
    p.acceptFailure = True
    self.assertEqual(mol.GetNumConformers(), 0)
    self.assertEqual(len(rdDepictor.GenerateDepictionMatching2DStructure(mol, template_ref,
      -1, None, p)), 0)
    self.assertEqual(mol.GetNumConformers(), 1)

  def testGenerateAlignedCoordsAlignOnly(self):
    template_ref_molblock = """
     RDKit          2D

  6  6  0  0  0  0  0  0  0  0999 V2000
  -13.7477    6.3036    0.0000 R#  0  0  0  0  0  0  0  0  0  0  0  0
  -13.7477    4.7567    0.0000 C   0  0  0  0  0  0  0  0  0  0  0  0
  -12.6540    3.6628    0.0000 C   0  0  0  0  0  0  0  0  0  0  0  0
  -13.7477    2.5691    0.0000 C   0  0  0  0  0  0  0  0  0  0  0  0
  -14.8414    3.6628    0.0000 N   0  0  0  0  0  0  0  0  0  0  0  0
  -11.1071    3.6628    0.0000 R#  0  0  0  0  0  0  0  0  0  0  0  0
  1  2  1  0
  2  3  1  0
  3  4  1  0
  4  5  1  0
  2  5  1  0
  3  6  1  0
M  RGP  2   1   1   6   2
M  END
"""
    template_ref = Chem.MolFromMolBlock(template_ref_molblock)
    mol_molblock = """
     RDKit          2D

 18 22  0  0  0  0  0  0  0  0999 V2000
    4.3922   -1.5699    0.0000 C   0  0  0  0  0  0  0  0  0  0  0  0
    2.9211   -2.0479    0.0000 C   0  0  0  0  0  0  0  0  0  0  0  0
    2.5995   -0.5349    0.0000 C   0  0  0  0  0  0  0  0  0  0  0  0
    3.3731    0.8046    0.0000 C   0  0  0  0  0  0  0  0  0  0  0  0
    4.8441    1.2825    0.0000 C   0  0  0  0  0  0  0  0  0  0  0  0
    4.0704   -0.0568    0.0000 C   0  0  0  0  0  0  0  0  0  0  0  0
    1.8666    0.7748    0.0000 N   0  0  0  0  0  0  0  0  0  0  0  0
    0.7736   -0.3197    0.0000 C   0  0  0  0  0  0  0  0  0  0  0  0
    0.7749   -1.8666    0.0000 C   0  0  0  0  0  0  0  0  0  0  0  0
   -0.7718   -1.8679    0.0000 N   0  0  0  0  0  0  0  0  0  0  0  0
   -0.7731   -0.3208    0.0000 C   0  0  0  0  0  0  0  0  0  0  0  0
   -1.8679    0.7718    0.0000 N   0  0  0  0  0  0  0  0  0  0  0  0
   -4.0718   -0.0598    0.0000 C   0  0  0  0  0  0  0  0  0  0  0  0
   -4.3933   -1.5729    0.0000 C   0  0  0  0  0  0  0  0  0  0  0  0
   -2.9222   -2.0509    0.0000 C   0  0  0  0  0  0  0  0  0  0  0  0
   -2.6008   -0.5379    0.0000 C   0  0  0  0  0  0  0  0  0  0  0  0
   -3.3744    0.8016    0.0000 C   0  0  0  0  0  0  0  0  0  0  0  0
   -4.8454    1.2795    0.0000 C   0  0  0  0  0  0  0  0  0  0  0  0
  9 10  1  0
 11 10  1  0
 11  8  1  0
  8  9  1  0
  4  5  1  0
  6  5  1  0
  7  6  1  0
  3  4  1  0
  3  7  1  0
  1  6  1  0
  2  3  1  0
  2  1  1  0
 17 18  1  0
 13 18  1  0
 12 13  1  0
 16 17  1  0
 16 12  1  0
 14 13  1  0
 15 16  1  0
 15 14  1  0
 12 11  1  0
  8  7  1  0
M  END
"""
    mol = Chem.MolFromMolBlock(mol_molblock)
    bondLength11_12 = rdMolTransforms.GetBondLength(mol.GetConformer(), 11, 12)
    bondLength5_6 = rdMolTransforms.GetBondLength(mol.GetConformer(), 5, 6)
    self.assertLess(abs(bondLength11_12 - bondLength5_6), 1.e-4)
    self.assertGreater(bondLength11_12, 2.3)
    #for alignOnly in (False, True):
    for alignOnly in (True,):
      mol = Chem.MolFromMolBlock(mol_molblock)
      p = rdDepictor.ConstrainedDepictionParams()
      p.allowRGroups = True
      p.alignOnly = alignOnly
      res = rdDepictor.GenerateDepictionMatching2DStructure(mol, template_ref, params=p)
      expectedMolIndices = [11, 10, 7, 8, 9, 6]
      self.assertTrue(all(templateRefAtomIdx == i and molAtomIdx == expectedMolIndices[i]
                          for i, (templateRefAtomIdx, molAtomIdx) in enumerate(res)))
      self.assertEqual(Chem.MolToSmiles(mol), "C1CC2CCC1N2C1CNC1N1C2CCC1CC2")
      self.assertTrue(all((mol.GetConformer().GetAtomPosition(molAtomIdx)
                           - template_ref.GetConformer().GetAtomPosition(templateRefAtomIdx)).LengthSq() < 1.e-4
                           for templateRefAtomIdx, molAtomIdx in res))
      bondLengthAli11_12 = rdMolTransforms.GetBondLength(mol.GetConformer(), 11, 12)
      bondLengthAli5_6 = rdMolTransforms.GetBondLength(mol.GetConformer(), 5, 6)
      self.assertLess(abs(bondLengthAli11_12 - bondLengthAli5_6), 1.e-4)
      if alignOnly:
        self.assertGreater(bondLengthAli11_12, 2.3)
      else:
        self.assertLess(bondLengthAli11_12, 1.6)

  def testRGroupMatchHeavyHydroNoneCharged(self):
    templateRef = Chem.MolFromMolBlock("""
  MJ201100                      

  7  7  0  0  0  0  0  0  0  0999 V2000
   -0.5804    1.2045    0.0000 C   0  0  0  0  0  0  0  0  0  0  0  0
   -1.2948    0.7920    0.0000 C   0  0  0  0  0  0  0  0  0  0  0  0
   -1.2948   -0.0330    0.0000 C   0  0  0  0  0  0  0  0  0  0  0  0
   -0.5804   -0.4455    0.0000 C   0  0  0  0  0  0  0  0  0  0  0  0
    0.1340   -0.0330    0.0000 A   0  0  0  0  0  0  0  0  0  0  0  0
    0.1340    0.7920    0.0000 C   0  0  0  0  0  0  0  0  0  0  0  0
    0.8485   -0.4455    0.0000 R#  0  0  0  0  0  0  0  0  0  0  0  0
  1  2  2  0  0  0  0
  2  3  1  0  0  0  0
  3  4  2  0  0  0  0
  6  1  1  0  0  0  0
  4  5  1  0  0  0  0
  5  6  2  0  0  0  0
  5  7  1  0  0  0  0
M  RGP  1   7   1
M  END
""")
    p = rdDepictor.ConstrainedDepictionParams()
    p.allowRGroups = True
    for alignOnly in (True, False):
      p.alignOnly = alignOnly
      mol = Chem.MolFromSmiles("Cc1ccccc1")
      self.assertEqual(mol.GetNumAtoms(), 7)
      self.assertEqual(len(rdDepictor.GenerateDepictionMatching2DStructure(mol, templateRef, params=p)), 7)
      mol = Chem.MolFromSmiles("c1ccccc1")
      self.assertEqual(mol.GetNumAtoms(), 6)
      self.assertEqual(len(rdDepictor.GenerateDepictionMatching2DStructure(mol, templateRef, params=p)), 6)
      smilesParams = Chem.SmilesParserParams()
      smilesParams.removeHs = False
      mol = Chem.MolFromSmiles("[H]c1ccccc1", smilesParams)
      self.assertEqual(mol.GetNumAtoms(), 7)
      self.assertEqual(len(rdDepictor.GenerateDepictionMatching2DStructure(mol, templateRef, params=p)), 7)
      mol = Chem.MolFromSmiles("n1ccccc1")
      self.assertEqual(mol.GetNumAtoms(), 6)
      self.assertEqual(len(rdDepictor.GenerateDepictionMatching2DStructure(mol, templateRef, params=p)), 6)
      mol = Chem.MolFromSmiles("C[n+]1ccccc1")
      self.assertEqual(mol.GetNumAtoms(), 7)
      self.assertEqual(len(rdDepictor.GenerateDepictionMatching2DStructure(mol, templateRef, params=p)), 7)

if __name__ == '__main__':
  unittest.main()



"""unit testing code for 3D stuff

"""
import os
import random
import sys
import unittest

from rdkit import Chem, RDConfig
from rdkit.Chem import AllChem, TorsionFingerprints


class TestCase(unittest.TestCase):

  def testConformerRMS(self):
    m1 = Chem.MolFromSmiles('CNc(n2)nc(C)cc2Nc(cc34)ccc3[nH]nc4')
    cids = AllChem.EmbedMultipleConfs(m1, 2)

    m2 = Chem.MolFromSmiles('CNc(n2)nc(C)cc2Nc(cc34)ccc3[nH]nc4')
    m2.AddConformer(m1.GetConformer(id=1))

    # test that the prealigned flag is working
    rms1 = AllChem.GetConformerRMS(m1, 0, 1, prealigned=True)
    rms2 = AllChem.GetConformerRMS(m1, 0, 1, prealigned=False)
    self.assertTrue((rms1 > rms2))

    # test that RMS is the same as calculated by AlignMol()
    self.assertAlmostEqual(rms2, AllChem.GetBestRMS(m2, m1, 1, 0), 3)

    # the RMS with itself must be zero
    rms2 = AllChem.GetConformerRMS(m1, 0, 0, prealigned=True)
    self.assertAlmostEqual(rms2, 0.0, 4)

  def testConformerRMSMatrix(self):
    m1 = Chem.MolFromSmiles('CNc(n2)nc(C)cc2Nc(cc34)ccc3[nH]nc4')
    cids = AllChem.EmbedMultipleConfs(m1, 3)

    m2 = Chem.MolFromSmiles('CNc(n2)nc(C)cc2Nc(cc34)ccc3[nH]nc4')
    m2.AddConformer(m1.GetConformer(id=0))

    # test that the RMS matrix has the correct size
    rmat = AllChem.GetConformerRMSMatrix(m1)
    self.assertEqual(len(rmat), 3)

    # test that the elements are in the right order
    self.assertAlmostEqual(rmat[0], AllChem.GetBestRMS(m1, m2, 1, 0), 3)
    self.assertAlmostEqual(rmat[1], AllChem.GetBestRMS(m1, m2, 2, 0), 3)

    # test the prealigned option
    rmat2 = AllChem.GetConformerRMSMatrix(m1, prealigned=True)
    self.assertAlmostEqual(rmat[0], rmat2[0])

  def testTorsionFingerprints(self):
    # we use the xray structure from the paper (JCIM, 52, 1499, 2012): 1DWD
    refFile = os.path.join(RDConfig.RDCodeDir, 'Chem', 'test_data', '1DWD_ligand.pdb')
    ref = Chem.MolFromSmiles(
      'NC(=[NH2+])c1ccc(C[C@@H](NC(=O)CNS(=O)(=O)c2ccc3ccccc3c2)C(=O)N2CCCCC2)cc1')
    mol = Chem.MolFromPDBFile(refFile)
    mol = AllChem.AssignBondOrdersFromTemplate(ref, mol)

    # the torsion lists
    tors_list, tors_list_rings = TorsionFingerprints.CalculateTorsionLists(mol)
    self.assertEqual(len(tors_list), 11)
    self.assertEqual(len(tors_list_rings), 4)
    self.assertAlmostEqual(tors_list[-1][1], 180.0, 4)
    tors_list, tors_list_rings = TorsionFingerprints.CalculateTorsionLists(mol, maxDev='spec')
    self.assertAlmostEqual(tors_list[-1][1], 90.0, 4)
    self.assertRaises(ValueError, TorsionFingerprints.CalculateTorsionLists, mol, maxDev='test')
    tors_list, tors_list_rings = TorsionFingerprints.CalculateTorsionLists(mol, symmRadius=0)
    self.assertEqual(len(tors_list[0][0]), 2)

    # the weights
    weights = TorsionFingerprints.CalculateTorsionWeights(mol)
    self.assertAlmostEqual(weights[4], 1.0)
    self.assertEqual(len(weights), len(tors_list + tors_list_rings))
    weights = TorsionFingerprints.CalculateTorsionWeights(mol, 15, 14)
    self.assertAlmostEqual(weights[3], 1.0)
    self.assertRaises(ValueError, TorsionFingerprints.CalculateTorsionWeights, mol, 15, 3)

    # the torsion angles
    tors_list, tors_list_rings = TorsionFingerprints.CalculateTorsionLists(mol)
    torsions = TorsionFingerprints.CalculateTorsionAngles(mol, tors_list, tors_list_rings)
    self.assertEqual(len(weights), len(torsions))
    self.assertAlmostEqual(torsions[2][0][0], 232.5346, 4)

    # the torsion fingerprint deviation
    tfd = TorsionFingerprints.CalculateTFD(torsions, torsions)
    self.assertAlmostEqual(tfd, 0.0)
    refFile = os.path.join(RDConfig.RDCodeDir, 'Chem', 'test_data', '1PPC_ligand.pdb')
    mol2 = Chem.MolFromPDBFile(refFile)
    mol2 = AllChem.AssignBondOrdersFromTemplate(ref, mol2)
    torsions2 = TorsionFingerprints.CalculateTorsionAngles(mol2, tors_list, tors_list_rings)
    weights = TorsionFingerprints.CalculateTorsionWeights(mol)
    tfd = TorsionFingerprints.CalculateTFD(torsions, torsions2, weights=weights)
    self.assertAlmostEqual(tfd, 0.0691, 4)
    tfd = TorsionFingerprints.CalculateTFD(torsions, torsions2)
    self.assertAlmostEqual(tfd, 0.1115, 4)

    # the wrapper functions
    tfd = TorsionFingerprints.GetTFDBetweenMolecules(mol, mol2)
    self.assertAlmostEqual(tfd, 0.0691, 4)

    # exactly equivalent to the above since the mols each only have one conformer
    tfd = TorsionFingerprints.GetBestTFDBetweenMolecules(mol2, mol)
    self.assertAlmostEqual(tfd, 0.0691, 4)

    mol.AddConformer(mol2.GetConformer(), assignId=True)
    mol.AddConformer(mol2.GetConformer(), assignId=True)
    tfd = TorsionFingerprints.GetTFDBetweenConformers(mol, confIds1=[0], confIds2=[1, 2])
    self.assertEqual(len(tfd), 2)
    self.assertAlmostEqual(tfd[0], 0.0691, 4)

    tfdmat = TorsionFingerprints.GetTFDMatrix(mol)
    self.assertEqual(len(tfdmat), 3)

    tfd = TorsionFingerprints.GetBestTFDBetweenMolecules(mol2, mol)
    self.assertAlmostEqual(tfd, 0, 4)

  def testTorsionFingerprintsAtomReordering(self):
    # we use the xray structure from the paper (JCIM, 52, 1499, 2012): 1DWD
    refFile = os.path.join(RDConfig.RDCodeDir, 'Chem', 'test_data', '1DWD_ligand.pdb')
    ref = Chem.MolFromSmiles(
      'NC(=[NH2+])c1ccc(C[C@@H](NC(=O)CNS(=O)(=O)c2ccc3ccccc3c2)C(=O)N2CCCCC2)cc1')
    mol1 = Chem.MolFromPDBFile(refFile)
    mol1 = AllChem.AssignBondOrdersFromTemplate(ref, mol1)

    refFile = os.path.join(RDConfig.RDCodeDir, 'Chem', 'test_data', '1DWD_ligand_reordered.pdb')
    mol2 = Chem.MolFromPDBFile(refFile)
    mol2 = AllChem.AssignBondOrdersFromTemplate(ref, mol2)

    tfd = TorsionFingerprints.GetTFDBetweenMolecules(mol1, mol2)
    self.assertEqual(tfd, 0.0)

  def testTorsionFingerprintsColinearBonds(self):
    # test that single bonds adjacent to triple bonds are ignored
    mol = Chem.MolFromSmiles('CCC#CCC')
    tors_list, tors_list_rings = TorsionFingerprints.CalculateTorsionLists(
      mol, ignoreColinearBonds=True)
    self.assertEqual(len(tors_list), 0)
    weights = TorsionFingerprints.CalculateTorsionWeights(mol, ignoreColinearBonds=True)
    self.assertEqual(len(weights), 0)

    # test that they are not ignored, but alternative atoms searched for
    tors_list, tors_list_rings = TorsionFingerprints.CalculateTorsionLists(
      mol, ignoreColinearBonds=False)
    self.assertEqual(len(tors_list), 1)
    self.assertEqual(tors_list[0][0][0], (0, 1, 4, 5))
    weights = TorsionFingerprints.CalculateTorsionWeights(mol, ignoreColinearBonds=False)
    self.assertEqual(len(weights), 1)

    # test that single bonds adjacent to terminal triple bonds are always ignored
    mol = Chem.MolFromSmiles('C#CCC')
    tors_list, tors_list_rings = TorsionFingerprints.CalculateTorsionLists(
      mol, ignoreColinearBonds=True)
    self.assertEqual(len(tors_list), 0)
    tors_list, tors_list_rings = TorsionFingerprints.CalculateTorsionLists(
      mol, ignoreColinearBonds=False)
    self.assertEqual(len(tors_list), 0)

  def testGithub4720(self):
    # exceptions with highly-coordinated atoms
    mol = Chem.MolFromSmiles('S(F)(F)(F)(F)(Cl)c1ccccc1')
    tors_list, tors_list_rings = TorsionFingerprints.CalculateTorsionLists(mol)
    self.assertEqual(len(tors_list), 1)

  def testTorsionAngleLeargerThan14(self):
    # incorrect value from more than 15-membered ring
    mol = Chem.MolFromSmiles('C1' + 'C' * 13 + 'C1')
    tors_list, tors_list_rings = TorsionFingerprints.CalculateTorsionLists(mol)
    self.assertAlmostEqual(tors_list_rings[-1][1], 180.0, 4)

  def assertBondStereoRoundTrips(self, fname):
    path = os.path.join(RDConfig.RDCodeDir, 'Chem', 'test_data', fname)
    mol = Chem.MolFromMolFile(path)
    refSmiles = mol.GetProp("_Name")
    self.assertTrue(len(refSmiles) > 0)
    self.assertEqual(Chem.MolToSmiles(mol, isomericSmiles=True), refSmiles)

    # now test Chem.DetectBondStereoChemistry more directly by constructing the molecule from scratch
    oldconf = mol.GetConformer(0)
    newconf = Chem.Conformer(mol.GetNumAtoms())
    newmol = Chem.RWMol()

    for atm in mol.GetAtoms():
      ratm = Chem.Atom(atm.GetAtomicNum())
      ratm.SetFormalCharge(atm.GetFormalCharge())
      newmol.AddAtom(ratm)

      atomidx = atm.GetIdx()
      pos = oldconf.GetAtomPosition(atomidx)
      newconf.SetAtomPosition(atomidx, pos)

    for bnd in mol.GetBonds():
      newmol.AddBond(bnd.GetBeginAtomIdx(), bnd.GetEndAtomIdx(), Chem.BondType(bnd.GetBondType()))
    newmol.AddConformer(newconf)

    Chem.SanitizeMol(newmol)
    Chem.DetectBondStereoChemistry(newmol, newmol.GetConformer())

    # these aren't necessary for this specific test case, but are for
    # a more general conversion routine, so would like to see them
    # tested eventually
    # Chem.AssignAtomChiralTagsFromStructure(newmol)
    # Chem.AssignStereochemistry(newmol)

    self.assertEqual(Chem.MolToSmiles(newmol, isomericSmiles=True), refSmiles)

  def testDetectBondStereoChemistry(self):
    self.assertBondStereoRoundTrips('cis.sdf')
    self.assertBondStereoRoundTrips('trans.sdf')

  def testEnumerateStereoisomersBasic(self):
    mol = Chem.MolFromSmiles('CC(F)=CC(Cl)C')
    smiles = set(
      Chem.MolToSmiles(i, isomericSmiles=True) for i in AllChem.EnumerateStereoisomers(mol))
    print(smiles)
    self.assertEqual(len(smiles), 4)

  def testEnumerateStereoisomersLargeRandomSample(self):
    # near max number of stereo centers allowed
    mol = Chem.MolFromSmiles('CC(F)=CC(Cl)C' * 31)
    smiles = set(
      Chem.MolToSmiles(i, isomericSmiles=True) for i in AllChem.EnumerateStereoisomers(mol))
    self.assertEqual(len(smiles), 1024)

  def testEnumerateStereoisomersWithCrazyNumberOfCenters(self):
    # insanely large numbers of isomers aren't a problem
    mol = Chem.MolFromSmiles('CC(F)=CC(Cl)C' * 101)
    opts = AllChem.StereoEnumerationOptions(rand=None, maxIsomers=13)
    smiles = set(
      Chem.MolToSmiles(i, isomericSmiles=True) for i in AllChem.EnumerateStereoisomers(mol, opts))
    self.assertEqual(len(smiles), 13)

  def testEnumerateStereoisomersRandomSamplingShouldBeDeterministicAndPortable(self):
    mol = Chem.MolFromSmiles('CC(F)=CC(Cl)C=C(Br)C(I)N')
    opts = AllChem.StereoEnumerationOptions(maxIsomers=2)
    smiles = set(
      Chem.MolToSmiles(i, isomericSmiles=True) for i in AllChem.EnumerateStereoisomers(mol, opts))
    # Python 3.8 switch how tuples were hashed, so we get different results there
    if sys.version_info < (3, 8):
      expected = set([
        'C/C(F)=C/[C@@H](Cl)/C=C(/Br)[C@@H](N)I',
        'C/C(F)=C/[C@H](Cl)/C=C(\\Br)[C@H](N)I',
      ])
    else:
      expected = set(
        ['C/C(F)=C\\[C@H](Cl)/C=C(\\Br)[C@H](N)I', 'C/C(F)=C/[C@@H](Cl)/C=C(\\Br)[C@H](N)I'])
    self.assertEqual(smiles, expected)

  def testEnumerateStereoisomersMaxIsomersShouldBeReturnedEvenWithTryEmbedding(self):
    m = Chem.MolFromSmiles('BrC=CC1OC(C2)(F)C2(Cl)C1')
    opts = AllChem.StereoEnumerationOptions(tryEmbedding=True, maxIsomers=8)
    isomers = set()
    for x in AllChem.EnumerateStereoisomers(m, options=opts):
      isomers.add(Chem.MolToSmiles(x, isomericSmiles=True))
    self.assertEqual(len(isomers), 8)

  def testEnumerateStereoisomersTryEmbeddingShouldNotInfiniteLoopWhenMaxIsomersIsLargerThanActual(
      self):
    m = Chem.MolFromSmiles('BrC=CC1OC(C2)(F)C2(Cl)C1')
    opts = AllChem.StereoEnumerationOptions(tryEmbedding=True, maxIsomers=1024)
    isomers = set()
    for x in AllChem.EnumerateStereoisomers(m, options=opts):
      isomers.add(Chem.MolToSmiles(x, isomericSmiles=True))
    self.assertEqual(len(isomers), 8)

  def testEnumerateStereoisomersRandomSeeding(self):
    opts = AllChem.StereoEnumerationOptions(rand=None, maxIsomers=3)
    mol = Chem.MolFromSmiles('CC(F)=CC(Cl)C')
    smiles = set(
      Chem.MolToSmiles(i, isomericSmiles=True) for i in AllChem.EnumerateStereoisomers(mol, opts))
    self.assertEqual(smiles,
                     set(['C/C(F)=C/[C@@H](C)Cl', 'C/C(F)=C\\[C@H](C)Cl', 'C/C(F)=C\\[C@@H](C)Cl']))

    opts = AllChem.StereoEnumerationOptions(rand=0xDEADBEEF)
    mol = Chem.MolFromSmiles('c1ccc2c(c1)C(=O)N(C2=O)C3CCC(=O)NC3=O')
    smiles = set(
      Chem.MolToSmiles(i, isomericSmiles=True) for i in AllChem.EnumerateStereoisomers(mol, opts))
    self.assertEqual(
      smiles,
      set(['O=C1CC[C@@H](N2C(=O)c3ccccc3C2=O)C(=O)N1', 'O=C1CC[C@H](N2C(=O)c3ccccc3C2=O)C(=O)N1']))

    class DeterministicRandom(random.Random):

      def __init__(self):
        random.Random.__init__(self)
        self.count = 0

      def getrandbits(self, n_bits):
        c = self.count
        self.count += 1
        return c

    rand = DeterministicRandom()
    opts = AllChem.StereoEnumerationOptions(rand=rand, maxIsomers=3)
    mol = Chem.MolFromSmiles('CCCC(=C(CCl)C(C)CBr)[C@H](F)C(C)C')
    smiles = [
      Chem.MolToSmiles(i, isomericSmiles=True) for i in AllChem.EnumerateStereoisomers(mol, opts)
    ]
    self.assertEqual(smiles, [
      'CCC/C(=C(\\CCl)[C@H](C)CBr)[C@H](F)C(C)C', 'CCC/C(=C(\\CCl)[C@@H](C)CBr)[C@H](F)C(C)C',
      'CCC/C(=C(/CCl)[C@H](C)CBr)[C@H](F)C(C)C'
    ])

  def testEnumerateStereoisomersOnlyUnassigned(self):
    # shouldn't enumerate anything
    fully_assigned = Chem.MolFromSmiles('C/C(F)=C/[C@@H](C)Cl')
    smiles = set(
      Chem.MolToSmiles(i, isomericSmiles=True)
      for i in AllChem.EnumerateStereoisomers(fully_assigned))
    self.assertEqual(smiles, set(['C/C(F)=C/[C@@H](C)Cl']))

    # should only enumerate the bond stereo
    partially_assigned = Chem.MolFromSmiles('CC(F)=C[C@@H](C)Cl')
    smiles = set(
      Chem.MolToSmiles(i, isomericSmiles=True)
      for i in AllChem.EnumerateStereoisomers(partially_assigned))
    self.assertEqual(smiles, set(['C/C(F)=C/[C@@H](C)Cl', 'C/C(F)=C\\[C@@H](C)Cl']))

    # should enumerate everything
    opts = AllChem.StereoEnumerationOptions(onlyUnassigned=False)
    smiles = set(
      Chem.MolToSmiles(i, isomericSmiles=True)
      for i in AllChem.EnumerateStereoisomers(fully_assigned, opts))
    print(smiles)
    self.assertEqual(
      smiles,
      set([
        'C/C(F)=C\\[C@@H](C)Cl',
        'C/C(F)=C\\[C@H](C)Cl',
        'C/C(F)=C/[C@H](C)Cl',
        'C/C(F)=C/[C@@H](C)Cl',
      ]))

  def testEnumerateStereoisomersOnlyUnique(self):
    mol = Chem.MolFromSmiles('FC(Cl)C(Cl)F')
    opts = AllChem.StereoEnumerationOptions(unique=False)
    smiles = [
      Chem.MolToSmiles(i, isomericSmiles=True) for i in AllChem.EnumerateStereoisomers(mol, opts)
    ]
    self.assertEqual(len(smiles), 4)
    print(set(smiles))
    self.assertEqual(len(set(smiles)), 3)

    mol = Chem.MolFromSmiles('FC(Cl)C(Cl)F')
    opts = AllChem.StereoEnumerationOptions(unique=True)
    smiles = set(
      Chem.MolToSmiles(i, isomericSmiles=True) for i in AllChem.EnumerateStereoisomers(mol, opts))
    self.assertEqual(
      smiles, set(['F[C@@H](Cl)[C@@H](F)Cl', 'F[C@H](Cl)[C@@H](F)Cl', 'F[C@H](Cl)[C@H](F)Cl']))

    mol = Chem.MolFromSmiles('CC=CC=CC')
    opts = AllChem.StereoEnumerationOptions(unique=False)
    smiles = [
      Chem.MolToSmiles(i, isomericSmiles=True) for i in AllChem.EnumerateStereoisomers(mol, opts)
    ]
    self.assertEqual(len(smiles), 4)
    self.assertEqual(len(set(smiles)), 3)

    mol = Chem.MolFromSmiles('CC=CC=CC')
    opts = AllChem.StereoEnumerationOptions(unique=True)
    smiles = set(
      Chem.MolToSmiles(i, isomericSmiles=True) for i in AllChem.EnumerateStereoisomers(mol, opts))
    self.assertEqual(smiles, set(['C/C=C/C=C/C', 'C/C=C\\C=C\\C', 'C/C=C\\C=C/C']))

    mol = Chem.MolFromSmiles('FC(Cl)C=CC=CC(F)Cl')
    opts = AllChem.StereoEnumerationOptions(unique=False)
    smiles = [
      Chem.MolToSmiles(i, isomericSmiles=True) for i in AllChem.EnumerateStereoisomers(mol, opts)
    ]
    self.assertEqual(len(smiles), 16)
    self.assertEqual(len(set(smiles)), 10)

    mol = Chem.MolFromSmiles('FC(Cl)C=CC=CC(F)Cl')
    opts = AllChem.StereoEnumerationOptions(unique=True)
    smiles = set(
      sorted(
        Chem.MolToSmiles(i, isomericSmiles=True)
        for i in AllChem.EnumerateStereoisomers(mol, opts)))
    self.assertEqual(
      smiles,
      set(
        sorted([
          'F[C@H](Cl)/C=C/C=C/[C@@H](F)Cl', 'F[C@H](Cl)/C=C/C=C/[C@H](F)Cl',
          'F[C@@H](Cl)/C=C/C=C/[C@@H](F)Cl', 'F[C@@H](Cl)/C=C\\C=C\\[C@@H](F)Cl',
          'F[C@H](Cl)/C=C\\C=C/[C@@H](F)Cl', 'F[C@H](Cl)/C=C/C=C\\[C@@H](F)Cl',
          'F[C@H](Cl)/C=C\\C=C\\[C@H](F)Cl', 'F[C@@H](Cl)/C=C\\C=C/[C@@H](F)Cl',
          'F[C@H](Cl)/C=C\\C=C/[C@H](F)Cl', 'F[C@H](Cl)/C=C\\C=C\\[C@@H](F)Cl'
        ])))

  def testEnumerateStereoisomersOnlyEnhancedStereo(self):
    rdbase = os.environ["RDBASE"]
    filename = os.path.join(rdbase, 'Code/GraphMol/FileParsers/test_data/two_centers_or.mol')
    mol = Chem.MolFromMolFile(filename)
    smiles = set(Chem.MolToSmiles(m) for m in AllChem.EnumerateStereoisomers(mol))
    # switches the centers linked by an "OR", but not the absolute group
    self.assertEqual(smiles, {r'C[C@H]([C@@H](C)F)[C@@H](C)Br', r'C[C@@H]([C@H](C)F)[C@@H](C)Br'})

    # we need the SMILES without the influence of the stereo groups:
    mol = Chem.RWMol(mol)
    mol.SetStereoGroups([])
    original_smiles = Chem.MolToSmiles(mol)
    self.assertIn(original_smiles, smiles)

  def testNoExtrasIfEnumeratingAllWithEnhancedStereo(self):
    """
    If the onlyUnassigned option is False, make sure that enhanced stereo
    groups aren't double-counted.
    """
    rdbase = os.environ["RDBASE"]
    filename = os.path.join(rdbase, 'Code/GraphMol/FileParsers/test_data/two_centers_or.mol')
    mol = Chem.MolFromMolFile(filename)

    opts = AllChem.StereoEnumerationOptions(onlyUnassigned=False, unique=False)
    smiles = [Chem.MolToSmiles(m) for m in AllChem.EnumerateStereoisomers(mol, opts)]
    print('!!!!', smiles)
    self.assertEqual(len(smiles), len(set(smiles)))
    self.assertEqual(len(smiles), 2**3)

  def testIssue2890(self):
    mol = Chem.MolFromSmiles('CC=CC')
    mol.GetBondWithIdx(1).SetStereo(Chem.rdchem.BondStereo.STEREOANY)

    self.assertEqual(len(list(AllChem.EnumerateStereoisomers(mol))), 2)

  def testIssue3231(self):
    mol = Chem.MolFromSmiles(
      'C[C@H](OC1=C(N)N=CC(C2=CN(C3C[C@H](C)NCC3)N=C2)=C1)C4=C(Cl)C=CC(F)=C4Cl')
    Chem.AssignStereochemistry(mol, force=True, flagPossibleStereoCenters=True)
    l = Chem.FindMolChiralCenters(mol, includeUnassigned=True)
    self.assertEqual(l, [(1, 'S'), (12, '?'), (14, 'S')])
    enumsi_opt = AllChem.StereoEnumerationOptions(maxIsomers=20, onlyUnassigned=False)
    isomers = list(AllChem.EnumerateStereoisomers(mol, enumsi_opt))
    chi_cents = []
    for iso in isomers:
      Chem.AssignStereochemistry(iso)
      chi_cents.append(Chem.FindMolChiralCenters(iso, includeUnassigned=True))
    self.assertEqual(sorted(chi_cents),
                     [[(1, 'R'), (12, 'R'),
                       (14, 'R')], [(1, 'R'), (12, 'R'),
                                    (14, 'S')], [(1, 'R'), (12, 'S'),
                                                 (14, 'R')], [(1, 'R'), (12, 'S'), (14, 'S')],
                      [(1, 'S'), (12, 'R'),
                       (14, 'R')], [(1, 'S'), (12, 'R'),
                                    (14, 'S')], [(1, 'S'), (12, 'S'),
                                                 (14, 'R')], [(1, 'S'), (12, 'S'), (14, 'S')]])

  def testIssue3505(self):
    m = Chem.MolFromSmiles('CCC(C)Br')
    mols = list(AllChem.EnumerateStereoisomers(m))
    self.assertEqual(len(mols), 2)
    for mol in mols:
      at = mol.GetAtomWithIdx(2)
      self.assertIn(at.GetChiralTag(),
                    [Chem.ChiralType.CHI_TETRAHEDRAL_CW, Chem.ChiralType.CHI_TETRAHEDRAL_CCW])
      self.assertTrue(at.HasProp("_ChiralityPossible"))

  def testEnumerateEitherDoubleStereo(self):
    """ EnumerateStereoisomers from MOL with explicit either cis/trans bond """
    rdbase = os.environ["RDBASE"]
    filename = os.path.join(rdbase, 'Code/GraphMol/FileParsers/test_data/simple_either.mol')
    mol = Chem.MolFromMolFile(filename)
    smiles = [Chem.MolToSmiles(m) for m in AllChem.EnumerateStereoisomers(mol)]
    self.assertEqual(set(smiles), {"C/C=C/C", "C/C=C\\C"})

  def testTryEmbeddingManyChirals(self):
    smiles = "C1" + "C(Cl)(Br)" * 40 + "C1"
    mol = Chem.MolFromSmiles(smiles)
    opts = AllChem.StereoEnumerationOptions(tryEmbedding=True, maxIsomers=2)
    self.assertEqual(len(list(AllChem.EnumerateStereoisomers(mol, options=opts))), 2)

  def testGithub6045(self):
    mol = Chem.MolFromSmiles('O[C@H](Br)[C@H](F)C |&1:1,3|')
    prods = list(AllChem.EnumerateStereoisomers(mol))
    self.assertEqual(len(prods), 2)
    for prod in prods:
      self.assertEqual(len(prod.GetStereoGroups()), 0)


if __name__ == '__main__':
  unittest.main()



# $Id$
#
#  Copyright (C) 2002-2006  greg Landrum and Rational Discovery LLC
#
#   @@ All Rights Reserved @@
#  This file is part of the RDKit.
#  The contents are covered by the terms of the BSD license
#  which is included in the file license.txt, found at the root
#  of the RDKit source tree.
#
"""Command line tool to construct an enrichment plot from saved composite models

Usage:  EnrichPlot [optional args] -d dbname -t tablename <models>

Required Arguments:
  -d "dbName": the name of the database for screening

  -t "tablename": provide the name of the table with the data to be screened

  <models>: file name(s) of pickled composite model(s).
     If the -p argument is also provided (see below), this argument is ignored.

Optional Arguments:
  - -a "list": the list of result codes to be considered active.  This will be
        eval'ed, so be sure that it evaluates as a list or sequence of
        integers. For example, -a "[1,2]" will consider activity values 1 and 2
        to be active

  - --enrich "list": identical to the -a argument above.

  - --thresh: sets a threshold for the plot.  If the confidence falls below
          this value, picking will be terminated

  - -H: screen only the hold out set (works only if a version of
        BuildComposite more recent than 1.2.2 was used).

  - -T: screen only the training set (works only if a version of
        BuildComposite more recent than 1.2.2 was used).

  - -S: shuffle activity values before screening

  - -R: randomize activity values before screening

  - -F *filter frac*: filters the data before training to change the
     distribution of activity values in the training set.  *filter frac*
     is the fraction of the training set that should have the target value.
     **See note in BuildComposite help about data filtering**

  - -v *filter value*: filters the data before training to change the
     distribution of activity values in the training set. *filter value*
     is the target value to use in filtering.
     **See note in BuildComposite help about data filtering**

  - -p "tableName": provides the name of a db table containing the
      models to be screened.  If you use this argument, you should also
      use the -N argument (below) to specify a note value.

  - -N "note": provides a note to be used to pull models from a db table.

  - --plotFile "filename": writes the data to an output text file (filename.dat)
    and creates a gnuplot input file (filename.gnu) to plot it

  - --showPlot: causes the gnuplot plot constructed using --plotFile to be
    displayed in gnuplot.

"""
from warnings import warn

warn('This module is deprecated and will be removed in the 2024.03 release', DeprecationWarning,
     stacklevel=2)
# from rdkit.Dbase.DbConnection import DbConnect

import pickle
import sys

import numpy

from rdkit import DataStructs, RDConfig
from rdkit.Dbase.DbConnection import DbConnect
from rdkit.ML import CompositeRun
from rdkit.ML.Data import DataUtils, SplitData, Stats

__VERSION_STRING = "2.4.0"


def cmp(t1, t2):
  return (t1 < t2) * -1 or (t1 > t2) * 1


def message(msg, noRet=0, dest=sys.stderr):
  """ emits messages to _sys.stderr_
      override this in modules which import this one to redirect output

      **Arguments**

        - msg: the string to be displayed

    """
  if noRet:
    dest.write('%s ' % (msg))
  else:
    dest.write('%s\n' % (msg))


def error(msg, dest=sys.stderr):
  """ emits messages to _sys.stderr_
      override this in modules which import this one to redirect output

      **Arguments**

        - msg: the string to be displayed

    """
  sys.stderr.write('ERROR: %s\n' % (msg))


def ScreenModel(mdl, descs, data, picking=[1], indices=[], errorEstimate=0):
  """ collects the results of screening an individual composite model that match
      a particular value

     **Arguments**

       - mdl: the composite model

       - descs: a list of descriptor names corresponding to the data set

       - data: the data set, a list of points to be screened.

       - picking: (Optional) a list of values that are to be collected.
         For examples, if you want an enrichment plot for picking the values
         1 and 2, you'd having picking=[1,2].

      **Returns**

        a list of 4-tuples containing:

           - the id of the point

           - the true result (from the data set)

           - the predicted result

           - the confidence value for the prediction

    """
  mdl.SetInputOrder(descs)

  for j in range(len(mdl)):
    tmp = mdl.GetModel(j)
    if hasattr(tmp, '_trainIndices') and not isinstance(tmp._trainIndices, dict):
      tis = {}
      if hasattr(tmp, '_trainIndices'):
        for v in tmp._trainIndices:
          tis[v] = 1
      tmp._trainIndices = tis

  res = []
  if mdl.GetQuantBounds():
    needsQuant = 1
  else:
    needsQuant = 0

  if not indices:
    indices = list(range(len(data)))
  nTrueActives = 0
  for i in indices:
    if errorEstimate:
      use = []
      for j in range(len(mdl)):
        tmp = mdl.GetModel(j)
        if not tmp._trainIndices.get(i, 0):
          use.append(j)
    else:
      use = None
    pt = data[i]
    pred, conf = mdl.ClassifyExample(pt, onlyModels=use)
    if needsQuant:
      pt = mdl.QuantizeActivity(pt[:])
    trueRes = pt[-1]
    if trueRes in picking:
      nTrueActives += 1
    if pred in picking:
      res.append((pt[0], trueRes, pred, conf))
  return nTrueActives, res


def AccumulateCounts(predictions, thresh=0, sortIt=1):
  """  Accumulates the data for the enrichment plot for a single model

      **Arguments**

        - predictions: a list of 3-tuples (as returned by _ScreenModels_)

        - thresh: a threshold for the confidence level.  Anything below
          this threshold will not be considered

        - sortIt: toggles sorting on confidence levels


      **Returns**

        - a list of 3-tuples:

          - the id of the active picked here

          - num actives found so far

          - number of picks made so far

    """
  if sortIt:
    predictions.sort(lambda x, y: cmp(y[3], x[3]))
  res = []
  nCorrect = 0
  nPts = 0
  for i in range(len(predictions)):
    ID, real, pred, conf = predictions[i]
    if conf > thresh:
      if pred == real:
        nCorrect += 1
      nPts += 1
      res.append((ID, nCorrect, nPts))

  return res


def MakePlot(details, final, counts, pickVects, nModels, nTrueActs=-1):
  if not hasattr(details, 'plotFile') or not details.plotFile:
    return

  dataFileName = '%s.dat' % (details.plotFile)
  outF = open(dataFileName, 'w+')
  i = 0
  while i < len(final) and counts[i] != 0:
    if nModels > 1:
      _, sd = Stats.MeanAndDev(pickVects[i])
      confInterval = Stats.GetConfidenceInterval(sd, len(pickVects[i]), level=90)
      outF.write('%d %f %f %d %f\n' %
                 (i + 1, final[i][0] / counts[i], final[i][1] / counts[i], counts[i], confInterval))
    else:
      outF.write('%d %f %f %d\n' %
                 (i + 1, final[i][0] / counts[i], final[i][1] / counts[i], counts[i]))
    i += 1
  outF.close()
  plotFileName = '%s.gnu' % (details.plotFile)
  gnuF = open(plotFileName, 'w+')
  gnuHdr = """# Generated by EnrichPlot.py version: %s
  set size square 0.7
  set xr [0:]
  set data styl points
  set ylab 'Num Correct Picks'
  set xlab 'Num Picks'
  set grid
  set nokey
  set term postscript enh color solid "Helvetica" 16
  set term X
  """ % (__VERSION_STRING)
  print(gnuHdr, file=gnuF)
  if nTrueActs > 0:
    print('set yr [0:%d]' % nTrueActs, file=gnuF)
  print('plot x with lines', file=gnuF)
  if nModels > 1:
    everyGap = i / 20
    print('replot "%s" using 1:2 with lines,' % (dataFileName), end='', file=gnuF)
    print('"%s" every %d using 1:2:5 with yerrorbars' % (dataFileName, everyGap), file=gnuF)
  else:
    print('replot "%s" with points' % (dataFileName), file=gnuF)
  gnuF.close()

  if hasattr(details, 'showPlot') and details.showPlot:
    try:
      from Gnuplot import Gnuplot
      p = Gnuplot()
      p('load "%s"' % (plotFileName))
      input('press return to continue...\n')
    except Exception:
      import traceback
      traceback.print_exc()


def Usage():
  """ displays a usage message and exits """
  sys.stderr.write(__doc__)
  sys.exit(-1)


if __name__ == '__main__':
  import getopt
  try:
    args, extras = getopt.getopt(sys.argv[1:], 'd:t:a:N:p:cSTHF:v:',
                                 ('thresh=', 'plotFile=', 'showPlot', 'pickleCol=', 'OOB', 'noSort',
                                  'pickBase=', 'doROC', 'rocThresh=', 'enrich='))
  except Exception:
    import traceback
    traceback.print_exc()
    Usage()

  details = CompositeRun.CompositeRun()
  CompositeRun.SetDefaults(details)

  details.activeTgt = [1]
  details.doTraining = 0
  details.doHoldout = 0
  details.dbTableName = ''
  details.plotFile = ''
  details.showPlot = 0
  details.pickleCol = -1
  details.errorEstimate = 0
  details.sortIt = 1
  details.pickBase = ''
  details.doROC = 0
  details.rocThresh = -1
  for arg, val in args:
    if arg == '-d':
      details.dbName = val
    if arg == '-t':
      details.dbTableName = val
    elif arg == '-a' or arg == '--enrich':
      details.activeTgt = eval(val)
      if not isinstance(details.activeTgt, (tuple, list)):
        # if (type(details.activeTgt) not in (types.TupleType, types.ListType)):
        details.activeTgt = (details.activeTgt, )

    elif arg == '--thresh':
      details.threshold = float(val)
    elif arg == '-N':
      details.note = val
    elif arg == '-p':
      details.persistTblName = val
    elif arg == '-S':
      details.shuffleActivities = 1
    elif arg == '-H':
      details.doTraining = 0
      details.doHoldout = 1
    elif arg == '-T':
      details.doTraining = 1
      details.doHoldout = 0
    elif arg == '-F':
      details.filterFrac = float(val)
    elif arg == '-v':
      details.filterVal = float(val)
    elif arg == '--plotFile':
      details.plotFile = val
    elif arg == '--showPlot':
      details.showPlot = 1
    elif arg == '--pickleCol':
      details.pickleCol = int(val) - 1
    elif arg == '--OOB':
      details.errorEstimate = 1
    elif arg == '--noSort':
      details.sortIt = 0
    elif arg == '--doROC':
      details.doROC = 1
    elif arg == '--rocThresh':
      details.rocThresh = int(val)
    elif arg == '--pickBase':
      details.pickBase = val

  if not details.dbName or not details.dbTableName:
    Usage()
    print('*******Please provide both the -d and -t arguments')

  message('Building Data set\n')
  dataSet = DataUtils.DBToData(details.dbName, details.dbTableName, user=RDConfig.defaultDBUser,
                               password=RDConfig.defaultDBPassword, pickleCol=details.pickleCol,
                               pickleClass=DataStructs.ExplicitBitVect)

  descs = dataSet.GetVarNames()
  nPts = dataSet.GetNPts()
  message('npts: %d\n' % (nPts))
  final = numpy.zeros((nPts, 2), float)
  counts = numpy.zeros(nPts, numpy.integer)
  selPts = [None] * nPts

  models = []
  if details.persistTblName:
    conn = DbConnect(details.dbName, details.persistTblName)
    message('-> Retrieving models from database')
    curs = conn.GetCursor()
    curs.execute("select model from %s where note='%s'" % (details.persistTblName, details.note))
    message('-> Reconstructing models')
    try:
      blob = curs.fetchone()
    except Exception:
      blob = None
    while blob:
      message(' Building model %d' % len(models))
      blob = blob[0]
      try:
        models.append(pickle.loads(str(blob)))
      except Exception:
        import traceback
        traceback.print_exc()
        print('Model failed')
      else:
        message('  <-Done')
      try:
        blob = curs.fetchone()
      except Exception:
        blob = None
    curs = None
  else:
    for modelName in extras:
      try:
        model = pickle.load(open(modelName, 'rb'))
      except Exception:
        import traceback
        print('problems with model %s:' % modelName)
        traceback.print_exc()
      else:
        models.append(model)
  nModels = len(models)
  pickVects = {}
  halfwayPts = [1e8] * len(models)
  for whichModel, model in enumerate(models):
    tmpD = dataSet
    try:
      seed = model._randomSeed
    except AttributeError:
      pass
    else:
      DataUtils.InitRandomNumbers(seed)
    if details.shuffleActivities:
      DataUtils.RandomizeActivities(tmpD, shuffle=1)
    if hasattr(model, '_splitFrac') and (details.doHoldout or details.doTraining):
      trainIdx, testIdx = SplitData.SplitIndices(tmpD.GetNPts(), model._splitFrac, silent=1)
      if details.filterFrac != 0.0:
        trainFilt, temp = DataUtils.FilterData(tmpD, details.filterVal, details.filterFrac, -1,
                                               indicesToUse=trainIdx, indicesOnly=1)
        testIdx += temp
        trainIdx = trainFilt
      if details.doTraining:
        testIdx, trainIdx = trainIdx, testIdx
    else:
      testIdx = list(range(tmpD.GetNPts()))

    message('screening %d examples' % (len(testIdx)))
    nTrueActives, screenRes = ScreenModel(model, descs, tmpD, picking=details.activeTgt,
                                          indices=testIdx, errorEstimate=details.errorEstimate)
    message('accumulating')
    runningCounts = AccumulateCounts(screenRes, sortIt=details.sortIt, thresh=details.threshold)
    if details.pickBase:
      pickFile = open('%s.%d.picks' % (details.pickBase, whichModel + 1), 'w+')
    else:
      pickFile = None

    for i, entry in enumerate(runningCounts):
      entry = runningCounts[i]
      selPts[i] = entry[0]
      final[i][0] += entry[1]
      final[i][1] += entry[2]
      v = pickVects.get(i, [])
      v.append(entry[1])
      pickVects[i] = v
      counts[i] += 1
      if pickFile:
        pickFile.write('%s\n' % (entry[0]))
      if entry[1] >= nTrueActives / 2 and entry[2] < halfwayPts[whichModel]:
        halfwayPts[whichModel] = entry[2]
    message('Halfway point: %d\n' % halfwayPts[whichModel])

  if details.plotFile:
    MakePlot(details, final, counts, pickVects, nModels, nTrueActs=nTrueActives)
  else:
    if nModels > 1:
      print('#Index\tAvg_num_correct\tConf90Pct\tAvg_num_picked\tNum_picks\tlast_selection')
    else:
      print('#Index\tAvg_num_correct\tAvg_num_picked\tNum_picks\tlast_selection')

    i = 0
    while i < nPts and counts[i] != 0:
      if nModels > 1:
        mean, sd = Stats.MeanAndDev(pickVects[i])
        confInterval = Stats.GetConfidenceInterval(sd, len(pickVects[i]), level=90)
        print('%d\t%f\t%f\t%f\t%d\t%s' % (i + 1, final[i][0] / counts[i], confInterval,
                                          final[i][1] / counts[i], counts[i], str(selPts[i])))
      else:
        print('%d\t%f\t%f\t%d\t%s' %
              (i + 1, final[i][0] / counts[i], final[i][1] / counts[i], counts[i], str(selPts[i])))
      i += 1

  mean, sd = Stats.MeanAndDev(halfwayPts)
  print('Halfway point: %.2f(%.2f)' % (mean, sd))



#
#  Copyright (C) 2004-2008  greg Landrum and Rational Discovery LLC
#
#   @@ All Rights Reserved @@
#  This file is part of the RDKit.
#  The contents are covered by the terms of the BSD license
#  which is included in the file license.txt, found at the root
#  of the RDKit source tree.
#

import unittest

from rdkit.Chem.Pharm3D.ExcludedVolume import ExcludedVolume


class TestCase(unittest.TestCase):

  def test_ExcludedVolume(self):
    # featInfo must have a length
    self.assertRaises(ValueError, ExcludedVolume, 123)
    self.assertRaises(ValueError, ExcludedVolume, [])
    self.assertRaises(ValueError, ExcludedVolume, [
      123,
    ])
    self.assertRaises(ValueError, ExcludedVolume, [
      [],
    ])

    featInfo = ([(0, ), 0.5, 1.0], )
    excludedVolume = ExcludedVolume(featInfo)
    self.assertEqual(excludedVolume.featInfo, featInfo)
    self.assertAlmostEqual(excludedVolume.exclusionDist, 3.0)
    excludedVolume = ExcludedVolume(featInfo, exclusionDist=3.14)
    self.assertEqual(excludedVolume.featInfo, featInfo)
    self.assertAlmostEqual(excludedVolume.exclusionDist, 3.14)


if __name__ == '__main__':  # pragma: nocover
  unittest.main()



import os
import sys
import tempfile
import urllib

from mod_python import apache
from WebUtils import General

General._version = "1.0.0"


def page(req, smiles='', width=300, height=300, highlight='[]', numbers=0, **kwargs):
  req.content_type = 'text/html'
  page = General.ConstructHtmlHeader(title='RD Depict')

  if smiles:
    uSmiles = urllib.quote_plus(smiles)
    uHighlight = urllib.quote_plus(highlight)
    url = '/RDExtras/MolImage.py/svg?smiles=%s&height=%s&width=%s&highlight=%s&numbers=%s' % (
      uSmiles, height, width, uHighlight, numbers)
    imgUrl = '/RDExtras/MolImage.py/gif?smiles=%s&height=%s&width=%s&useCactvs=1' % (uSmiles,
                                                                                     height, width)

    page += """<center>
    <table>
    <tr>
      <td><embed src="%s" Name="SVGEmbed" width=%s height=%s type="image/svg-xml">
      <td><img src="%s" width=%s height=%s></td>
    </tr>
    </table>
    </center>""" % (url, width, height, imgUrl, width, height)
    page += '<center><b>SMILES:</b>%s</center>' % smiles
  if not smiles:
    smiles = '""'
  if not numbers or numbers == 'off':
    checked = ""
  else:
    checked = "checked"
  page += """<center>
  <form action="/RDExtras/MolDepict.py/page">
  <input type=text name="smiles" value=%s size=80>
  <input type=submit Value="Depict">
  <br><input type=checkbox name="numbers" %s> Numbers
  <input type=hidden name="width" value=%s>
  <input type=hidden name="height" value=%s>
  <input type=hidden name="highlight" value=%s>
  </center>
  """ % (smiles, checked, width, height, highlight)

  page += General.ConstructHtmlFooter(includeRestart=0, logoutText='')

  return page



#!/usr/bin/env python
# quick script for comparing SVGs in 2 different directories.
import argparse
import glob
from pathlib import Path

parser = argparse.ArgumentParser(description='Make an HTML table for comparing images.')
parser.add_argument('--dir1', required=True, help='Name of first directory.')
parser.add_argument('--dir2', required=True, help='Name of second directory.')
parser.add_argument('--file-glob', required=True, help='glob for files in first directory')
parser.add_argument('--outfile', default='side_by_side.html', help='Name of HTML file')

args = parser.parse_args()

d1 = args.dir1
d2 = args.dir2

if not Path(d1).exists():
    print(f'Directory {d1} missing.')
    exit(1)
if not Path(d2).exists():
    print(f'Directory {d2} missing.')
    exit(1)

with open(args.outfile, 'w') as f:
    f.write(f'''<!DOCTYPE html>
<html lang="en">
  <head>
    <meta charset="UTF-8" />
    <meta name="viewport" content="width=device-width, initial-scale=1.0" />
    <meta http-equiv="X-UA-Compatible" content="ie=edge" />
    <title>{d1} vs {d2}</title>
    <link href="index.css" rel="stylesheet" />
  </head>
  <body style="background-color:#eeeeee;">
    <table border="1">
      <tr>
        <td></td>
        <td>{d1}</td>
        <td>{d2}</td>
      </tr>\n''')

    inglob = Path(d1)
    fns = [Path(fn) for fn in inglob.glob(args.file_glob)]
    fns.sort(key=lambda f: f.stat().st_mtime, reverse=True)

    for fp in fns:
        fn = fp.name
        if not fn.endswith('.svg') and not fn.endswith('.png'):
            continue
        fns = fn.replace('.svg', '')
        f.write(f'''      <tr>
        <td>{fns}</td>
        <td><img src="{d1}/{fn}" alt="{fns}"/></td>
        <td><img src="{d2}/{fn}" alt="{fns}"/></td>
        </tr>\n''')

    f.write('''    </table>
  </body>
</html>\n''')
